
import sys
import re
import math
import string
import itertools
from copy import copy, deepcopy
import warnings
import importlib.resources

from ..utils import MM, rotate_point, bbox_intersect
from .primitives import *
from ..graphic_objects import Region, Line
from ..apertures import RectangleAperture, CircleAperture, ApertureMacroInstance
from ..aperture_macros.parse import ApertureMacro, ParameterExpression, VariableExpression
from ..aperture_macros import primitive as amp
from .kicad import footprints as kfp
from . import data as package_data


class ProtoBoard(Board):
    def __init__(self, w, h, content, margin=None, corner_radius=None, mounting_hole_dia=None, mounting_hole_offset=None, unit=MM):
        corner_radius = corner_radius or unit(1.5, MM)
        super().__init__(w, h, corner_radius, unit=unit)
        self.margin = margin or unit(2, MM)
        self.content = content

        if mounting_hole_dia:
            mounting_hole_offset = mounting_hole_offset or mounting_hole_dia*2
            ko = mounting_hole_offset + mounting_hole_dia*(0.5 + 0.25)

            stack = MechanicalHoleStack(mounting_hole_dia, unit=unit)
            self.mounting_holes = [
                    Pad(mounting_hole_offset, mounting_hole_offset, pad_stack=stack, unit=unit),
                    Pad(w-mounting_hole_offset, mounting_hole_offset, pad_stack=stack, unit=unit),
                    Pad(mounting_hole_offset, h-mounting_hole_offset, pad_stack=stack, unit=unit),
                    Pad(w-mounting_hole_offset, h-mounting_hole_offset, pad_stack=stack, unit=unit)]
            for hole in self.mounting_holes:
                self.add(hole)

            self.keepouts.append(((0, 0), (ko, ko)))
            self.keepouts.append(((w-ko, 0), (w, ko)))
            self.keepouts.append(((0, h-ko), (ko, h)))
            self.keepouts.append(((w-ko, h-ko), (w, h)))

        self.generate()

    def generate(self, unit=MM):
        bbox = ((self.margin, self.margin), (self.w-self.margin, self.h-self.margin))
        bbox = unit.convert_bounds_from(self.unit, bbox)
        hole_bboxes = [hole.bounding_box(unit) for hole in self.mounting_holes]
        for obj in self.content.generate(bbox, (True, True, True, True), self.keepouts, self.margin, True, unit):
            if isinstance(obj, Text):
                # It's okay for the text to go into the mounting hole keepouts, we just don't want it to overlap with
                # the actual mounting holes.
                if not any(bbox_intersect(obj.bounding_box(unit), hole_bbox) for hole_bbox in hole_bboxes):
                    self.add(obj, keepout_errors='ignore')
            else:
                self.add(obj, keepout_errors='ignore')


class PropLayout:
    def __init__(self, content, direction, proportions):
        self.content = list(content)
        if direction not in ('h', 'v'):
            raise ValueError('direction must be one of "h", or "v".')
        self.direction = direction
        self.proportions = list(proportions)
        if len(content) != len(proportions):
            raise ValueError('proportions and content must have same length')

    def increment_x(self):
        if self.direction == 'h':
            return 0
        else:
            return max(obj.increment_x() for obj in self.content) 

    def increment_y(self):
        if self.direction == 'v':
            return 0
        else:
            return max(obj.increment_y() for obj in self.content) 

    def generate(self, bbox, border_text, keepouts, text_margin, two_sided, unit=MM):
        for i, (bbox, child) in enumerate(self.layout_2d(bbox, unit)):
            first = bool(i == 0)
            last = bool(i == len(self.content)-1)
            yield from child.generate(bbox, (
                border_text[0] and (last or self.direction == 'h'),
                border_text[1] and (last or self.direction == 'v'),
                border_text[2] and (first or self.direction == 'h'),
                border_text[3] and (first or self.direction == 'v'),
                ), keepouts, text_margin, two_sided, unit)

    def fit_size(self, w, h, unit=MM):
        widths = []
        heights = []
        for ((x_min, y_min), (x_max, y_max)), child in self.layout_2d(((0, 0), (w, h)), unit):
            if not isinstance(child, EmptyProtoArea):
                widths.append(x_max - x_min)
                heights.append(y_max - y_min)
        if self.direction == 'h':
            return sum(widths), max(heights, default=0)
        else:
            return max(widths, default=0), sum(heights)

    def layout_2d(self, bbox, unit=MM):
        (x, y), (w, h) = bbox
        w, h = w-x, h-y
        total_w, total_h = w, h

        actual_l = 0
        target_l = 0
        total_l = total_w if self.direction == 'h' else total_h

        sizes = []
        for l, child in zip(self.layout(w if self.direction == 'h' else h, unit), self.content):
            this_x, this_y = x, y
            this_w, this_h = w, h
            target_l += l

            if self.direction == 'h':
                this_w = target_l - actual_l
            else:
                this_h = target_l - actual_l

            this_w, this_h = child.fit_size(this_w, this_h, unit)

            if self.direction == 'h':
                x += this_w
                actual_l += this_w
                this_h = h
            else:
                y += this_h
                actual_l += this_h
                this_w = w

            sizes.append(((this_x, this_y), (this_w, this_h)))

        # We don't want to pull in a whole bin packing implementation here, but we also don't want to be too dumb. Thus,
        # we just take the leftover space and distribute it to the children in descending increment (grid / pitch size).
        children_sorted = reversed(sorted(enumerate(self.content),
                                  key=lambda e: e[1].increment_x() if self.direction == 'h' else e[1].increment_y()))

        excess_l = total_l - actual_l
        children_extra = [0] * len(self.content)
        for child_i, child in children_sorted:
            increment = child.increment_x() if self.direction=='h' else child.increment_y()
            adjustment = increment * (excess_l//increment) if increment > 0 else excess_l
            children_extra[child_i] += adjustment
            excess_l -= adjustment

        adjust_l = 0
        for extra, ((this_x, this_y), (this_w, this_h)), child in zip(children_extra, sizes, self.content):
            if self.direction == 'h':
                this_x += adjust_l
                this_w += extra
            else:
                this_y += adjust_l
                this_h += extra
            adjust_l += extra

            yield ((this_x, this_y), (this_x+this_w, this_y+this_h)), child


    def layout(self, length, unit=MM):
        out = [ eval_value(value, MM(length, unit)) for value in self.proportions ]
        total_length = sum(value for value in out if value is not None)
        if length - total_length < -1e-6:
            raise ValueError(f'Proportions sum to {total_length} mm, which is greater than the available space of {length} mm.')

        leftover = length - total_length
        sum_props = sum( (value or 1.0) for value in self.proportions if not isinstance(value, str) )
        return [ unit(leftover * (value or 1.0) / sum_props if not isinstance(value, str) else calculated, MM)
                for value, calculated in zip(self.proportions, out) ]

    @property
    def single_sided(self):
        return all(elem.single_sided for elem in self.content)

    def __str__(self):
        children = ', '.join( f'{elem}:{width}' for elem, width in zip(self.content, self.proportions))
        return f'PropLayout[{self.direction.upper()}]({children})'


class TwoSideLayout:
    def __init__(self, top, bottom):
        self.top, self.bottom = top, bottom

        if not top.single_sided or not bottom.single_sided:
            warnings.warn('Two-sided pattern used on one side of a TwoSideLayout')

    def increment_x(self):
        return max(self.top.increment_x(), self.bottom.increment_x())

    def increment_y(self):
        return max(self.top.increment_y(), self.bottom.increment_y())

    def fit_size(self, w, h, unit=MM):
        w1, h1 = self.top.fit_size(w, h, unit)
        w2, h2 = self.bottom.fit_size(w, h, unit)
        if isinstance(self.top, EmptyProtoArea):
            if isinstance(self.bottom, EmptyProtoArea):
                return w1, h1
            return w2, h2
        if isinstance(self.bottom, EmptyProtoArea):
            return w1, h1
        return max(w1, w2), max(h1, h2)

    def generate(self, bbox, border_text, keepouts, text_margin, two_sided, unit=MM):
        yield from self.top.generate(bbox, border_text, keepouts, text_margin, False, unit)
        for obj in self.bottom.generate(bbox, border_text, keepouts, text_margin, False, unit):
            obj.flip = not obj.flip
            yield obj


def numeric(start=1):
    def gen():
        nonlocal start
        for i in itertools.count(start):
            yield str(i)

    return gen


def alphabetic(case='upper', alphabet=None):
    if case not in ('lower', 'upper'):
        raise ValueError('case must be one of "lower" or "upper".')

    if alphabet is None:
        index = string.ascii_lowercase if case == 'lower' else string.ascii_uppercase
    else:
        index = alphabet
    n = len(index)

    def gen():
        nonlocal index, n

        for i in itertools.count():
            if i<n:
                yield index[i]
                continue

            i -= n
            if i<n*n:
                yield index[i//n] + index[i%n]
                continue

            i -= n*n
            if i<n*n*n:
                yield index[i//(n*n)] + index[(i//n)%n] + index[i%n]

            else:
                raise ValueError(f'row/column index {i} out of range {n**3 + n**2 + n}')

    return gen


@dataclass
class BreadboardArea:
    drill: float = 0.9
    clearance: float = 0.5
    signal_trace_width: float = 0.8
    power_trace_width: float = 1.5
    pitch_x: float = 2.54
    pitch_y: float = 2.54
    power_rail_pitch: float = 2.54
    power_rail_space: float = 2.54
    num_power_rails: int = 2
    num_holes: int = 5
    center_space: float = 5.08
    horizontal: bool = True
    margin: float = 0
    font_size: float = 1.0
    font_stroke: float = 0.2
    unit: object = MM

    def fit_size(self, w, h, unit=MM):
        m = unit(self.margin, self.unit)
        w = max(0, w-2*m)
        h = max(0, h-2*m)
        
        pitch_x = self.pitch_x
        pitch_y = self.pitch_y
        if self.horizontal:
            pitch_x, pitch_y = pitch_y, pitch_x
        
        w_mod = round((w + 5e-7) % unit(pitch_x, self.unit), 6)
        h_mod = round((h + 5e-7) % unit(pitch_y, self.unit), 6)
        w_fit, h_fit = round(w - w_mod, 6), round(h - h_mod, 6)
        return w_fit + 2*m, h_fit + 2*m

    @property
    def width_across(self):
        w = self.pitch_x * self.num_holes * 2 + self.center_space
        if self.num_power_rails > 0:
            # include one power rail pitch unit for the space between adjacent tiles.
            w += 2*self.power_rail_space + (2*self.num_power_rails-1) * self.power_rail_pitch
        return w

    def increment_x(self):
        if self.horizontal:
            return self.pitch_y
        else:
            return self.width_across

    def increment_y(self):
        if self.horizontal:
            return self.width_across
        else:
            return self.pitch_y

    @property
    def single_sided(self):
        return False

    def generate(self, bbox, border_text, keepouts, text_margin, two_sided, unit=MM):
        (x, y), (w, h) = self.unit.convert_bounds_from(unit, bbox)
        w, h = w-x-self.margin, h-y-self.margin
        ox, oy = (y, x) if self.horizontal else (x, y)

        signal_ap = CircleAperture(self.signal_trace_width, unit=self.unit)
        power_ap = CircleAperture(self.power_trace_width, unit=self.unit)

        pad_dia = min(self.pitch_x, self.pitch_y) - self.clearance
        tht_pad = THTPad.circle(self.drill, pad_dia)

        available_width = h if self.horizontal else w
        length_along = w if self.horizontal else h

        # Key:
        #   H - signal pad
        #   C - center space
        #   P - power pad
        #   R - power rail space

        pitch_key = {
                'H': self.pitch_x,
                'C': self.center_space,
                'P': self.power_rail_pitch,
                'R': self.power_rail_space}

        layouts = []

        for i in range(self.num_holes):
            sig = 'H' * (i+1)
            layouts.append(sig)

        layouts.append(f'{sig}C{sig}')

        for i in range(self.num_power_rails):
            pwr = 'P' * (i+1)
            layouts.append(f'{pwr}R{sig}C{sig}')
            layouts.append(f'{pwr}R{sig}C{sig}R{pwr}')

        while len(layouts[-1]) <= available_width // self.pitch_x:
            pre = layouts[-1]

            for i in range(self.num_holes):
                sig = 'H' * (i+1)
                layouts.append(f'{pre}R{sig}')

            for i in range(self.num_holes):
                sig = 'H' * (i+1)
                for i in range(self.num_power_rails):
                    pwr = 'P' * (i+1)
                    layouts.append(f'{pre}R{sig}C{sig}R{pwr}')
                layouts.append(f'{pre}R{sig}R{pwr}')

            layouts.append(f'{pre}R{sig}C{sig}')

            for i in range(self.num_power_rails):
                pwr = 'P' * (i+1)
                layouts.append(f'{pre}R{sig}C{sig}R{pwr}')

        best_layout, leftover_space = None, None
        for layout in layouts:
            actual_width = sum(pitch_key[e] for e in layout)

            if actual_width <= available_width:
                best_layout = layout
                leftover_space = available_width - actual_width

        if best_layout is None:
            return # We don't have enough space to do anything
        print(f'Chosen layout: {best_layout} with {leftover_space} left over')

        rail_start = {}
        rail_end = {}
        n_y = round(length_along//self.pitch_y)
        for j in range(n_y):
            y = oy + self.margin + self.pitch_y*(j + 0.5) + (length_along - (n_y*self.pitch_y))/2
            pos_across = ox + self.margin + leftover_space/2
            last_e = 'R'
            for e, group in itertools.groupby(enumerate(best_layout), key=lambda e: e[1]):
                group = list(group)
                num = len(group)
                local_pitch = pitch_key[e]
                
                points = []
                for k, _e in group:
                    x = pos_across + local_pitch/2
                    ax, ay = (y, x) if self.horizontal else (x, y)
                    px, py = (self.pitch_y, local_pitch) if self.horizontal else (local_pitch, self.pitch_y)

                    if not any(bbox_intersect(ko, ((ax-px/2, ay-py/2), (ax+px/2, ay+py/2))) for ko in keepouts):
                        points.append((ax, ay))

                        if e == 'H':
                            yield Pad(ax, ay, pad_stack=tht_pad, unit=self.unit)

                        elif e == 'P':
                            yield Pad(ax, ay, pad_stack=tht_pad, unit=self.unit)

                        if k not in rail_start:
                            rail_start[k] = (ax, ay)
                        rail_end[k] = (ax, ay)

                    pos_across += local_pitch

                if e == 'H':
                    if len(points) > 1:
                        yield Trace(self.signal_trace_width, points[0], points[-1], unit=self.unit)

                    label = f'{j+1}'

                    if last_e == 'R':
                        if points:
                            tx, ty = points[0]

                            if self.horizontal:
                                ty -= self.pitch_x/2
                                yield Text(tx, ty, label, self.font_size, self.font_stroke, 'center', 'top', unit=self.unit)
                                yield Text(tx, ty, label, self.font_size, self.font_stroke, 'center', 'top', unit=self.unit, flip=True)
                            else:
                                tx -= self.pitch_x/2
                                yield Text(tx, ty, label, self.font_size, self.font_stroke, 'right', 'middle', unit=self.unit)
                                yield Text(tx, ty, label, self.font_size, self.font_stroke, 'right', 'middle', unit=self.unit, flip=True)

                    else:
                        if points:
                            tx, ty = points[-1]

                            if self.horizontal:
                                ty += self.pitch_x/2
                                yield Text(tx, ty, label, self.font_size, self.font_stroke, 'center', 'bottom', unit=self.unit)
                                yield Text(tx, ty, label, self.font_size, self.font_stroke, 'center', 'bottom', unit=self.unit, flip=True)
                            else:
                                tx += self.pitch_x/2
                                yield Text(tx, ty, label, self.font_size, self.font_stroke, 'left', 'middle', unit=self.unit)
                                yield Text(tx, ty, label, self.font_size, self.font_stroke, 'left', 'middle', unit=self.unit, flip=True)
                last_e = e

        if self.num_power_rails == 2 and best_layout.count('P') >= 2:
            power_rail_labels = ['-', '+'] * best_layout.count('P')
        else:
            power_rail_labels = [e for _, e in zip(best_layout, alphabetic(alphabet='ZXYWVU')())]
        signal_labels = alphabetic()() # yes, twice.

        line_ap = CircleAperture(self.power_trace_width, unit=self.unit)

        for i, e in enumerate(best_layout):
            start = rail_start.get(i)
            end = rail_end.get(i)

            if e == 'P':
                if start not in (None, end):
                    yield Trace(self.power_trace_width, start, end, unit=self.unit)
                    le_line = [Line(*start, *end, aperture=line_ap, unit=self.unit)]
                    yield Graphics(0, 0, top_silk=le_line, bottom_silk=le_line, unit=self.unit)

                label = power_rail_labels.pop(0)

            elif e == 'H':
                label = next(signal_labels)
            else:
                label = None

            if label:
                tx1, ty1 = start
                tx2, ty2 = end

                if self.horizontal:
                    tx1 -= self.pitch_y/2
                    tx2 += self.pitch_y/2

                    yield Text(tx1, ty1, label, self.font_size, self.font_stroke, 'right', 'middle', unit=self.unit)
                    yield Text(tx1, ty1, label, self.font_size, self.font_stroke, 'right', 'middle', unit=self.unit, flip=True)
                    yield Text(tx2, ty2, label, self.font_size, self.font_stroke, 'left', 'middle', unit=self.unit)
                    yield Text(tx2, ty2, label, self.font_size, self.font_stroke, 'left', 'middle', unit=self.unit, flip=True)

                else:
                    ty1 -= self.pitch_y/2
                    ty2 += self.pitch_y/2

                    yield Text(tx1, ty1, label, self.font_size, self.font_stroke, 'center', 'top', unit=self.unit)
                    yield Text(tx1, ty1, label, self.font_size, self.font_stroke, 'center', 'top', unit=self.unit, flip=True)
                    yield Text(tx2, ty2, label, self.font_size, self.font_stroke, 'center', 'bottom', unit=self.unit)
                    yield Text(tx2, ty2, label, self.font_size, self.font_stroke, 'center', 'bottom', unit=self.unit, flip=True)


class PatternProtoArea:
    def __init__(self, pitch_x, pitch_y=None, obj=None, numbers=True, font_size=None, font_stroke=None, number_x_gen=alphabetic(), number_y_gen=numeric(), interval_x=None, interval_y=None, margin=0, unit=MM):
        self.pitch_x = pitch_x
        self.pitch_y = pitch_y or pitch_x
        self.margin = margin
        self.obj = obj
        self.unit = unit
        self.numbers = numbers
        self.font_size = font_size or unit(1.0, MM)
        self.font_stroke = font_stroke or unit(0.2, MM)
        self.interval_x = interval_x
        self.interval_y = interval_y
        self.number_x_gen, self.number_y_gen = number_x_gen, number_y_gen

    def increment_x(self):
        return self.pitch_x

    def increment_y(self):
        return self.pitch_y

    def fit_size(self, w, h, unit=MM):
        (min_x, min_y), (max_x, max_y) = self.fit_rect(((0, 0), (max(0, w-2*self.margin), max(0, h-2*self.margin))))
        return max_x-min_x + 2*self.margin, max_y-min_y + 2*self.margin

    def fit_rect(self, bbox, unit=MM):
        (x, y), (w, h) = bbox
        w, h = w-x, h-y

        w_mod = round((w + 5e-7) % unit(self.pitch_x, self.unit), 6)
        h_mod = round((h + 5e-7) % unit(self.pitch_y, self.unit), 6)
        w_fit, h_fit = round(w - w_mod, 6), round(h - h_mod, 6)

        x = x + (w-w_fit)/2
        y = y + (h-h_fit)/2
        return (x, y), (x+w_fit, y+h_fit)

    def generate(self, bbox, border_text, keepouts, text_margin, two_sided, unit=MM):
        (x, y), (w, h) = bbox
        w, h = w-x, h-y

        n_x = int((w + 0.001)//unit(self.pitch_x, self.unit))
        n_y = int((h + 0.001)//unit(self.pitch_y, self.unit))
        off_x = (w - n_x*unit(self.pitch_x, self.unit)) / 2
        off_y = (h - n_y*unit(self.pitch_y, self.unit)) / 2

        if self.numbers:
            # Center row/column numbers in available margin. Note the swapped axes below - the Y (row) numbers are
            # centered in X direction, and vice versa.
            _idx, max_x_num = list(zip(range(n_x), self.number_x_gen()))[-1]
            _idx, max_y_num = list(zip(range(n_y), self.number_y_gen()))[-1]
            bbox_test_x = Text(0, 0, max_y_num, self.font_size, self.font_stroke, 'left', 'top', unit=self.unit)
            bbox_test_y = Text(0, 0, max_x_num, self.font_size, self.font_stroke, 'left', 'top', unit=self.unit)
            test_w = abs(bbox_test_x.bounding_box()[1][0] - bbox_test_x.bounding_box()[0][0])
            test_h = abs(bbox_test_y.bounding_box()[1][1] - bbox_test_y.bounding_box()[0][1])
            text_off_x = max(0, (off_x + text_margin - test_w)) / 2
            text_off_y = max(0, (off_y + text_margin - test_h)) / 2

            test_w = abs(bbox_test_y.bounding_box()[1][0] - bbox_test_y.bounding_box()[0][0])
            test_h = abs(bbox_test_x.bounding_box()[1][1] - bbox_test_x.bounding_box()[0][1])

            interval_x, interval_y = self.interval_x, self.interval_y
            if interval_x is None:
                interval_x = 1 if test_w < 0.8*self.pitch_x else 5
            if interval_y is None:
                interval_y = 1 if test_h < 0.8*self.pitch_y else 2

            for i, lno_i in list(zip(reversed(range(n_y)), self.number_y_gen())):
                if i == 0 or i == n_y - 1 or (i+1) % interval_y == 0:
                    t_y = off_y + y + (n_y - 1 - i + 0.5) * self.pitch_y

                    if border_text[3]:
                        t_x = x + off_x - text_off_x
                        yield Text(t_x, t_y, lno_i, self.font_size, self.font_stroke, 'right', 'middle', unit=self.unit)
                        if two_sided:
                            yield Text(t_x, t_y, lno_i, self.font_size, self.font_stroke, 'right', 'middle', flip=True, unit=self.unit)

                    if border_text[1]:
                        t_x = x + w - off_x + text_off_x
                        yield Text(t_x, t_y, lno_i, self.font_size, self.font_stroke, 'left', 'middle', unit=self.unit)
                        if two_sided:
                            yield Text(t_x, t_y, lno_i, self.font_size, self.font_stroke, 'left', 'middle', flip=True, unit=self.unit)

            for i, lno_i in zip(range(n_x), self.number_x_gen()):
                # We print every interval'th number, as well as the first and the last numbers.
                # The complex condition below is to avoid the corner case where interval is larger than 1, and the last
                # interval'th number is right next to the last number, and the two could overlap. In this case, we
                # suppress the last interval'th number.
                if i == 0 or i == n_x - 1 or ((i+1) % interval_x == 0 and (interval_x == 1 or i != n_x-2)):
                    t_x = off_x + x + (i + 0.5) * self.pitch_x

                    if border_text[2]:
                        t_y = y + off_y - text_off_y
                        yield Text(t_x, t_y, lno_i, self.font_size, self.font_stroke, 'center', 'top', unit=self.unit)
                        if two_sided:
                            yield Text(t_x, t_y, lno_i, self.font_size, self.font_stroke, 'center', 'top', flip=True, unit=self.unit)

                    if border_text[0]:
                        t_y = y + h - off_y + text_off_y
                        yield Text(t_x, t_y, lno_i, self.font_size, self.font_stroke, 'center', 'bottom', unit=self.unit)
                        if two_sided:
                            yield Text(t_x, t_y, lno_i, self.font_size, self.font_stroke, 'center', 'bottom', flip=True, unit=self.unit)


        for j in range(n_y):
            for i in range(n_x):
                x0 = off_x + x + i*self.pitch_x
                y0 = off_y + y + j*self.pitch_y
                x1 = x0 + self.pitch_x
                y1 = y0 + self.pitch_y

                border_n = (j == 0)     or any(bbox_intersect(ko, ((x0, y0-self.pitch_y), (x1, y0))) for ko in keepouts)
                border_s = (j == n_y-1) or any(bbox_intersect(ko, ((x0, y1), (x1, y1+self.pitch_y))) for ko in keepouts)
                border_w = (i == 0)     or any(bbox_intersect(ko, ((x0-self.pitch_x, y0), (x0, y1))) for ko in keepouts)
                border_e = (i == n_x-1) or any(bbox_intersect(ko, ((x1, y0), (x1+self.pitch_x, y1))) for ko in keepouts)
                border = (border_s, border_w, border_n, border_e)

                print({
                    (0, 0, 0, 0): '┼',
                    (1, 0, 0, 0): '┴',
                    (0, 1, 0, 0): '├',
                    (0, 0, 1, 0): '┬',
                    (0, 0, 0, 1): '┤',
                    (1, 1, 0, 0): '└',
                    (0, 1, 1, 0): '┌',
                    (0, 0, 1, 1): '┐',
                    (1, 0, 0, 1): '┘',
                    }.get(tuple(map(int, border)), '.'), end=('' if i < n_x-1 else '\n'))

                if any(bbox_intersect(ko, ((x0, y0), (x1, y1))) for ko in keepouts):
                    continue

                obj = self.obj
                if isinstance(obj, PadStack):
                    if hasattr(obj, 'grid_variant'):
                        obj = obj.grid_variant(i, j, border)
                        if obj is None:
                            continue

                    px = self.unit(off_x + x, unit) + (i + 0.5) * self.pitch_x
                    py = self.unit(off_y + y, unit) + (j + 0.5) * self.pitch_y
                    yield Pad(px, py, pad_stack=obj, unit=self.unit)
                    if two_sided and self.single_sided:
                        yield Pad(px, py, pad_stack=obj, flip=True, unit=self.unit)
                    continue

                elif hasattr(self.obj, 'inst'):
                    inst = self.obj.inst(i, j, border)
                    if not inst:
                        continue
                else:
                    inst = copy(self.obj)

                inst.x = inst.unit(off_x + x, unit) + (i + 0.5) * inst.unit(self.pitch_x, self.unit)
                inst.y = inst.unit(off_y + y, unit) + (j + 0.5) * inst.unit(self.pitch_y, self.unit)
                yield inst

                if two_sided and self.single_sided:
                    inst = copy(inst)
                    inst.flip = not inst.flip
                    yield inst

    @property
    def single_sided(self):
        return self.obj.single_sided


class EmptyProtoArea:
    def __init__(self, copper_fill=False):
        self.copper_fill = copper_fill

    def increment_x(self):
        return 0

    def increment_y(self):
        return 0

    def fit_size(self, w, h, unit=MM):
        return w, h

    def generate(self, bbox, border_text, keepouts, text_margin, two_sided, unit=MM):
        if self.copper_fill:
            (min_x, min_y), (max_x, max_y) = bbox
            group = ObjectGroup(0, 0, top_copper=[Region([(min_x, min_y), (max_x, min_y), (max_x, max_y), (min_x, max_y)],
                                                unit=unit, polarity_dark=True)])
            if two_sided:
                group.bottom_copper = group.top_copper
            group.bounding_box = lambda *args, **kwargs: None
            yield group

    @property
    def single_sided(self):
        return True


@dataclass(frozen=True, slots=True)
class ManhattanPads(PadStack):
    w: float = None
    h: float = None
    gap: float = 0.2

    @property
    def single_sided(self):
        return True

    @property
    def apertures(self):
        w = self.w
        h = self.h or w

        p = (w-2*self.gap)/2
        q = (h-2*self.gap)/2
        small_ap = RectangleAperture(p, q, unit=self.unit)

        s = min(w, h) / 2 / math.sqrt(2)
        large_ap = RectangleAperture(s, s, unit=self.unit).rotated(math.pi/4)
        large_ap_neg = RectangleAperture(s+2*self.gap, s+2*self.gap, unit=self.unit).rotated(math.pi/4)

        a = self.gap/2 + p/2
        b = self.gap/2 + q/2

        for layer in ('copper', 'mask'):
            yield PadStackAperture(small_ap, 'top', layer, -a, -b)
            yield PadStackAperture(small_ap, 'top', layer, -a,  b)
            yield PadStackAperture(small_ap, 'top', layer,  a, -b)
            yield PadStackAperture(small_ap, 'top', layer,  a,  b)
            yield PadStackAperture(large_ap_neg, 'top', layer,  0,  0, invert=True)
            yield PadStackAperture(large_ap, 'top', layer,  0,  0)


@dataclass(frozen=True, slots=True)
class RFGroundProto(PadStack):
    pitch: float = 2.54
    drill: float = 0.9
    clearance: float = 0.3
    via_drill: float = 0.4
    via_dia: float = 0.8
    pad_dia: float = None
    trace_width: float = None
    _: KW_ONLY = None
    suppress_via: bool = False

    @property
    def single_sided(self):
        return False

    @property
    def apertures(self):
        unit = self.unit
        pitch = self.pitch
        trace_width, pad_dia = self.trace_width, self.pad_dia

        if pad_dia is None:
            if trace_width is None:
                trace_width = 0.3
            pad_dia = pitch - trace_width - 2*self.clearance 
        elif trace_width is None:
            trace_width = pitch - pad_dia - 2*self.clearance

        via_ap = RectangleAperture(self.via_dia, self.via_dia, unit=unit).rotated(math.pi/4)
        pad_ap = CircleAperture(pad_dia, unit=unit)
        pad_neg_ap = CircleAperture(pad_dia+2*self.clearance, unit=unit)
        ground_ap = RectangleAperture(pitch + unit(0.01, MM), pitch + unit(0.01, MM), unit=unit)
        pad_drill = ExcellonTool(self.drill, plated=True, unit=unit)
        via_drill = ExcellonTool(self.via_drill, plated=True, unit=unit)

        for side in 'top', 'bottom':
            yield PadStackAperture(ground_ap, side, 'copper')
            yield PadStackAperture(pad_neg_ap, side, 'copper', invert=True)
            yield PadStackAperture(pad_ap, side, 'copper')
            yield PadStackAperture(pad_ap, side, 'mask')

            if not self.suppress_via:
                yield PadStackAperture(via_ap, side, 'copper', pitch/2, pitch/2)
                yield PadStackAperture(via_ap, side, 'mask', pitch/2, pitch/2)

        yield PadStackAperture(pad_drill, 'drill', 'plated')
        if not self.suppress_via:
            yield PadStackAperture(via_drill, 'drill', 'plated', pitch/2, pitch/2)

    def grid_variant(self, x, y, border):
        border_s, border_w, border_n, border_e = border
        if border_e or border_s:
            return replace(self, suppress_via=True)
        else:
            return self


@dataclass(frozen=True, slots=True)
class THTFlowerProto(PadStack):
    pitch: float = 2.54
    drill: float = 0.9
    diameter: float = 2.0
    clearance: float = 0.5
    border_s: bool = False
    border_w: bool = False
    border_n: bool = False
    border_e: bool = False

    @property
    def single_sided(self):
        return False

    @property
    def apertures(self):
        p = self.diameter / 2
        pad_dist_diag = math.sqrt(2) * (self.pitch - p) - self.drill
        pad_dist_ortho = 2*self.pitch - self.diameter - self.drill
        pad_dia = self.drill + max(0, min(pad_dist_diag, pad_dist_ortho) - self.clearance)

        pad = THTPad.circle(self.drill, pad_dia, paste=False, unit=self.unit)

        for ox, oy, brd in ((-p, 0, self.border_w), (p, 0, self.border_e), (0, -p, self.border_n), (0, p, self.border_s)):
            if not brd:
                for stack_ap in pad.apertures:
                    yield replace(stack_ap, offset_x=ox, offset_y=oy)

        middle_ap = CircleAperture(self.diameter, unit=self.unit)
        for side in ('top', 'bottom'):
            for layer in ('copper', 'mask'):
                yield PadStackAperture(middle_ap, side, layer)
    
    def grid_variant(self, x, y, border):
        border_s, border_w, border_n, border_e = border
        if ((x % 2 == 0) and (y % 2 == 0)) or ((x % 2 == 1) and (y % 2 == 1)):
            return replace(self, border_s=border_s, border_w=border_w, border_n=border_n, border_e=border_e)

        return None

#    def bounding_box(self, unit=MM):
#        x, y, rotation = self.abs_pos
#        p = self.pitch/2
#        return unit.convert_bounds_from(self.unit, ((x-p, y-p), (x+p, y+p)))

class PoweredProto(Graphics):
    """ Cell primitive for "powered" THT breadboards. This cell type is based on regular THT pads in a 100 mil grid, but
    adds small SMD pads diagonally between the THT pads. These SMD pads are interconnected with traces and vias in such
    a way that every second one is inter-linked, forming two fully connected grids. Next to every THT pad you have one
    pad of each grid, so this layout is awesome for distributing power throughout the board.

    This design is based on one that Yajima Manufacturing Akizuki Denshi, Akihabara's finest electronics store sells for
    next to nothing. Sadly, they don't ship internationally and they don't even have an English website, but if you ever
    are in Akihabara, Tokyo, Japan I can *highly* recommend a visit. The ones Yajima make are better than what this will
    produce since the Yajima ones use a two-colored silkscreen to visually distinguish the two power pad grids.

    Links:
    Akizuki Denshi product page: https://akizukidenshi.com/catalog/g/gP-07214/
    Yajima Manufacturing Corporation website: http://www.yajima-works.co.jp/index.html
    """

    @property
    def single_sided(self):
        return False

    def __init__(self, pitch=None, drill=None, clearance=None, power_pad_dia=None, via_size=None, trace_width=None, unit=MM):
        super().__init__(0, 0)
        self.unit = unit
        self.pitch = pitch = pitch or unit(2.54, MM)
        self.drill = drill = drill or unit(0.9, MM)
        self.clearance = clearance = clearance or unit(0.3, MM)
        self.trace_width = trace_width = trace_width or unit(0.3, MM)
        self.via_size = via_size = via_size or unit(0.4, MM)

        main_pad_dia = pitch - trace_width - 2*clearance
        power_pad_dia_max = math.sqrt(2)*pitch - main_pad_dia - 2*clearance
        if power_pad_dia is None:
            power_pad_dia = power_pad_dia_max - clearance # reduce some more to give the user more room
        elif power_pad_dia > power_pad_dia_max:
            warnings.warn(f'Power pad diameter {power_pad_dia} > {power_pad_dia_max} violates pad-to-pad clearance')
        self.power_pad_dia = power_pad_dia

        main_ap = CircleAperture(main_pad_dia, unit=unit)
        power_ap = CircleAperture(self.power_pad_dia, unit=unit)

        for l in [self.top_copper, self.bottom_copper]:
            l.append(Flash(0, 0, aperture=main_ap, unit=unit))

            l.append(Flash(-pitch/2, -pitch/2, aperture=power_ap, unit=unit))
            l.append(Flash(-pitch/2,  pitch/2, aperture=power_ap, unit=unit))
            l.append(Flash( pitch/2, -pitch/2, aperture=power_ap, unit=unit))
            l.append(Flash( pitch/2,  pitch/2, aperture=power_ap, unit=unit))

        self.drill_pth.append(Flash(0, 0, ExcellonTool(drill, plated=True, unit=unit), unit=unit))
        self.drill_pth.append(Flash(-pitch/2, -pitch/2, ExcellonTool(via_size, plated=True, unit=unit), unit=unit))

        self.top_mask = copy(self.top_copper)
        self.bottom_mask = copy(self.bottom_copper)

        self.line_ap = CircleAperture(trace_width, unit=unit)
        self.top_copper.append(Line(-pitch/2, -pitch/2, -pitch/2, pitch/2, aperture=self.line_ap, unit=unit))
        self.top_copper.append(Line(pitch/2, -pitch/2, pitch/2, pitch/2, aperture=self.line_ap, unit=unit))
        self.bottom_copper.append(Line(-pitch/2, -pitch/2, pitch/2, -pitch/2, aperture=self.line_ap, unit=unit))
        self.bottom_copper.append(Line(-pitch/2, pitch/2, pitch/2, pitch/2, aperture=self.line_ap, unit=unit))

    def inst(self, x, y, border):
        inst = copy(self)
        if (x + y) % 2 == 0:
            inst.drill_pth = inst.drill_pth[:-1]

        c = self.power_pad_dia/2 + self.clearance
        p = self.pitch/2

        if x == 1:
            inst.top_silk = [Line(-p, -p+c, -p, p-c, aperture=self.line_ap, unit=self.unit)]
        elif x % 2 == 0:
            inst.top_silk = [Line(p, -p+c, p, p-c, aperture=self.line_ap, unit=self.unit)]

        if y == 0:
            inst.bottom_silk = [Line(-p+c, -p, p-c, -p, aperture=self.line_ap, unit=self.unit)]
        elif y % 2 == 1:
            inst.bottom_silk = [Line(-p+c, p, p-c, p, aperture=self.line_ap, unit=self.unit)]

        return inst

    def bounding_box(self, unit=MM):
        x, y, rotation, flip = self.abs_pos
        p = self.pitch/2
        return unit.convert_bounds_from(self.unit, ((x-p, y-p), (x+p, y+p)))


class SpikyProto(ObjectGroup):
    """ Cell primitive for the "spiky" protoboard designed by @electroniceel and published on github at the URL below.
    This layout has small-ish standard THT pads, but in between these pads it puts a grid of SMD pads that are designed
    for easy solder bridging to allow for the construction of traces from solder bridging.

    Github URL: https://github.com/electroniceel/protoboard
    """

    def __init__(self, pitch=None, drill=None, clearance=None, power_pad_dia=None, via_size=None, trace_width=None, unit=MM):
        super().__init__(0, 0, unit=unit)
        res = importlib.resources.files(package_data)

        self.fp_center = kfp.Footprint.load(res.joinpath('center-pad-spikes.kicad_mod').read_text(encoding='utf-8'))
        self.corner_pad = kfp.FootprintInstance(1.27, 1.27, self.fp_center, unit=MM)

        self.pad = kfp.Footprint.load(res.joinpath('tht-0.8.kicad_mod').read_text(encoding='utf-8'))
        self.center_pad = kfp.FootprintInstance(0, 0, self.pad, unit=MM)

        self.fp_between = kfp.Footprint.load(res.joinpath('pad-between-spiked.kicad_mod').read_text(encoding='utf-8'))
        self.right_pad = kfp.FootprintInstance(1.27, 0, self.fp_between, unit=MM)
        self.top_pad = kfp.FootprintInstance(0, 1.27, self.fp_between, rotation=-math.pi/2, unit=MM)

    @property
    def objects(self):
        return [x for x in (self.center_pad, self.corner_pad, self.right_pad, self.top_pad) if x is not None]

    @objects.setter
    def objects(self, value):
        pass

    def inst(self, x, y, border):
        border_s, border_w, border_n, border_e = border
        inst = copy(self)

        if border_e:
            inst.corner_pad = inst.right_pad = None

        if border_s:
            inst.corner_pad = inst.top_pad = None

        return inst


@dataclass(frozen=True, slots=True)
class StarburstPad(PadStack):
    # Starburst pattern inspired by elecfreaks' "flower" protoboard
    pitch_x: float = 2.54
    pitch_y: float = 2.54
    trace_width_x: float = 1.4
    trace_width_y: float = 1.4
    solder_clearance: float = 0.4
    mask_width: float = 0.5
    drill: float = 0.9
    annular_ring: float = 1.2

    @property
    def apertures(self):
        var = ParameterExpression
        # parameters:  [1: pitch_x,
        #               2: trace_width_x,
        #               3: pitch_y,
        #               4: trace_width_y,
        #               5: diagonal_clearance,
        #               6: annular_ring_width]
        starburst_macro = ApertureMacro('STARB', 6, primitives=(
            amp.CenterLine(MM, 1, var(1), var(2)),
            amp.CenterLine(MM, 1, var(4), var(3)),
            amp.VectorLine(MM, 0, var(5), -var(1)/2, -var(3)/2, var(1)/2, var(3)/2),
            amp.VectorLine(MM, 0, var(5), var(1)/2, -var(3)/2, -var(1)/2, var(3)/2),
            amp.Circle(MM, 1, var(6)),
            ))

        main_ap = ApertureMacroInstance(starburst_macro, (self.pitch_x - self.solder_clearance, # 1
                                                          self.trace_width_x,                   # 2
                                                          self.pitch_y - self.solder_clearance, # 3
                                                          self.trace_width_y,                   # 4
                                                          self.mask_width,                      # 5
                                                          self.annular_ring), unit=self.unit)   # 6

        mask_ap = ApertureMacroInstance(starburst_macro, (self.pitch_x,                         # 1
                                                          self.trace_width_x,                   # 2
                                                          self.pitch_y,                         # 3
                                                          self.trace_width_y,                   # 4
                                                          self.mask_width,                      # 5
                                                          self.annular_ring), unit=self.unit)   # 6

        yield PadStackAperture(main_ap, 'top', 'copper')
        yield PadStackAperture(mask_ap, 'top', 'mask')
        yield PadStackAperture(main_ap, 'bottom', 'copper')
        yield PadStackAperture(mask_ap, 'bottom', 'mask')

        drill = ExcellonTool(self.drill, plated=True, unit=self.unit)
        yield PadStackAperture(drill, 'drill', 'plated', 0, 0)

class AlioCell(Positioned):
    """ Cell primitive for the ALio protoboard designed by arief ibrahim adha and published on hackaday.io at the URL
    below. Similar to electroniceel's spiky protoboard, this layout has small-ish standard THT pads, but in between
    these pads it puts a grid of SMD pads that are designed for easy solder bridging to allow for the construction of
    traces from solder bridging.

    Hackaday.io URL: https://hackaday.io/project/28570/
    """

    def __init__(self, pitch=None, drill=None, clearance=None, link_pad_width=None, link_trace_width=None, via_size=None, unit=MM):
        super().__init__(0, 0, unit=unit)
        self.pitch = pitch or unit(2.54, MM)
        self.drill = drill or unit(0.9, MM)
        self.clearance = clearance or unit(0.3, MM)
        self.link_pad_width = link_pad_width or unit(1.1, MM)
        self.link_trace_width = link_trace_width or unit(0.5, MM)
        self.via_size = via_size or unit(0.4, MM)
        self.border_s, self.border_w, self.border_n, self.border_e = False, False, False, False
        self.inst_x, self.inst_y = None, None

    @property
    def single_sided(self):
        return False

    def inst(self, x, y, border):
        inst = copy(self)
        inst.border_s, inst.border_w, inst.border_n, inst.border_e = border
        inst.inst_x, inst.inst_y = x, y
        return inst

    def bounding_box(self, unit):
        x, y, rotation, flip = self.abs_pos
        # FIXME hack
        return self.unit.convert_bounds_to(unit, ((x-self.pitch/2, y-self.pitch/2), (x+self.pitch/2, y+self.pitch/2)))

    def render(self, layer_stack, cache=None):
        x, y, rotation, flip = self.abs_pos
        def xf(fe):
            fe = copy(fe)
            fe.rotate(rotation)
            fe.offset(x, y, self.unit)
            return fe

        var = ParameterExpression
        foo = VariableExpression(var(2)/2 - var(1)/2 + var(4))
        bar = VariableExpression(var(4)+var(6))
        # parameters: [1: total height = pad width, 2: pitch, 3: trace width, 4: corner radius, 5: rotation, 6: clearance]
        alio_main_macro = ApertureMacro('ALIOM', 6, primitives=(
            amp.CenterLine(MM, 1, var(2)-var(6), var(2)-var(3)-2*var(6), 0, 0, var(5)),
            amp.Outline(MM, 0, 5, (
                -var(2)/2,          -var(2)/2,
                -var(2)/2,          -(foo-bar),
                -foo,               -(foo-bar),
                -(foo-bar),         -foo,
                -(foo-bar),         -var(2)/2, 
                -var(2)/2,          -var(2)/2,
                ), var(5)),
            amp.Outline(MM, 0, 5, (
                -var(2)/2,          var(2)/2,
                -var(2)/2,          (foo-bar),
                -foo,               (foo-bar),
                -(foo-bar),         foo,
                -(foo-bar),         var(2)/2, 
                -var(2)/2,          var(2)/2,
                ), var(5)),
            amp.Outline(MM, 0, 5, (
                 var(2)/2,          -var(2)/2,
                 var(2)/2,          -(foo-bar),
                 foo,               -(foo-bar),
                 (foo-bar),         -foo,
                 (foo-bar),         -var(2)/2, 
                 var(2)/2,          -var(2)/2,
                ), var(5)),
            amp.Outline(MM, 0, 5, (
                 var(2)/2,          var(2)/2,
                 var(2)/2,          (foo-bar),
                 foo,               (foo-bar),
                 (foo-bar),         foo,
                 (foo-bar),         var(2)/2, 
                 var(2)/2,          var(2)/2,
                ), var(5)),
            amp.Circle(MM, 0, 2*bar, -foo, -foo, var(5)),
            amp.Circle(MM, 0, 2*bar, -foo,  foo, var(5)),
            amp.Circle(MM, 0, 2*bar,  foo, -foo, var(5)),
            amp.Circle(MM, 0, 2*bar,  foo,  foo, var(5)),
            ))
        corner_radius = (self.link_pad_width - self.link_trace_width)/3
        main_ap = ApertureMacroInstance(alio_main_macro, (self.link_pad_width,         # 1
                                                          self.pitch,                  # 2
                                                          self.link_trace_width,       # 3
                                                          corner_radius,               # 4
                                                          rotation,                    # 5
                                                          self.clearance), unit=MM)    # 6
        main_ap_90 = ApertureMacroInstance(alio_main_macro, (self.link_pad_width,      # 1
                                                          self.pitch,                  # 2
                                                          self.link_trace_width,       # 3
                                                          corner_radius,               # 4
                                                          rotation-90,                 # 5
                                                          self.clearance), unit=MM)    # 6
        main_drill = ExcellonTool(self.drill, plated=True, unit=self.unit)
        via_drill = ExcellonTool(self.via_size, plated=True, unit=self.unit)

        # parameters: [1: total height = pad width, 2: total width, 3: trace width, 4: corner radius, 5: rotation]
        alio_macro = ApertureMacro('ALIOP', primitives=(
            amp.CenterLine(MM, 1, var(1)-2*var(4), var(1), 0, 0, var(5)),
            amp.CenterLine(MM, 1, var(1), var(1)-2*var(4), 0, 0, var(5)),
            amp.Circle(MM, 1, 2*var(4), -var(1)/2+var(4), -var(1)/2+var(4), var(5)),
            amp.Circle(MM, 1, 2*var(4), -var(1)/2+var(4),  var(1)/2-var(4), var(5)),
            amp.Circle(MM, 1, 2*var(4),  var(1)/2-var(4), -var(1)/2+var(4), var(5)),
            amp.Circle(MM, 1, 2*var(4),  var(1)/2-var(4),  var(1)/2-var(4), var(5)),
            amp.CenterLine(MM, 1, var(2), var(3), -var(2)/2 + var(1)/2, 0, var(5)),
            ))
        alio_dark = ApertureMacroInstance(alio_macro, (self.link_pad_width,         # 1
                                                       self.pitch-self.clearance,   # 2
                                                       self.link_trace_width,       # 3
                                                       corner_radius,               # 4
                                                       rotation), unit=MM)          # 5
        alio_dark_90 = ApertureMacroInstance(alio_macro, (self.link_pad_width,          # 1
                                                          self.pitch-self.clearance,    # 2
                                                          self.link_trace_width,        # 3
                                                          corner_radius,                # 4
                                                          rotation+90), unit=MM)        # 5

        end_pad = RectangleAperture(self.link_trace_width, self.pitch - 2*self.clearance - self.link_pad_width, unit=self.unit)
        end_pad_90 = end_pad.rotated(math.pi/2)

        # all layers are identical here
        for side, use in (('top', 'copper'), ('top', 'mask'), ('bottom', 'copper'), ('bottom', 'mask')):
            if side == 'top':
                layer_stack[side, use].objects.insert(0, xf(Flash(0, 0, aperture=main_ap, unit=self.unit)))
                if not self.border_s and not self.border_e:
                    layer_stack[side, use].objects.append(xf(Flash(self.pitch/2, self.pitch/2, aperture=alio_dark, unit=self.unit)))
                if self.border_e and not self.border_s:
                    layer_stack[side, use].objects.append(xf(Flash(0, self.pitch/2, aperture=end_pad_90, unit=self.unit)))
            else:
                layer_stack[side, use].objects.insert(0, xf(Flash(0, 0, aperture=main_ap_90, unit=self.unit)))
                if not self.border_e and not self.border_s:
                    layer_stack[side, use].objects.append(xf(Flash(self.pitch/2, self.pitch/2, aperture=alio_dark_90, unit=self.unit)))
                if self.border_s and not self.border_e:
                    layer_stack[side, use].objects.append(xf(Flash(self.pitch/2, 0, aperture=end_pad, unit=self.unit)))

        layer_stack.drill_pth.append(Flash(x, y, aperture=main_drill, unit=self.unit))
        if not (self.border_e or self.border_s):
            layer_stack.drill_pth.append(xf(Flash(self.pitch/2, self.pitch/2, aperture=via_drill, unit=self.unit)))


def convert_to_mm(value, unit):
    unitl  = unit.lower()
    if unitl == 'mm':
        return value
    elif unitl == 'cm':
        return value*10
    elif unitl == 'in':
        return value*25.4
    elif unitl == 'mil':
        return value/1000*25.4
    else:
        raise ValueError(f'Invalid unit {unit}, allowed units are mm, cm, in, and mil.')


_VALUE_RE = re.compile(r'([0-9]*\.?[0-9]+)(cm|mm|in|mil|%)')
def eval_value(value, total_length=None):
    if not isinstance(value, str):
        return None

    m = _VALUE_RE.match(value.lower())
    number, unit = m.groups()
    if unit == '%':
        if total_length is None:
            raise ValueError('Percentages are not allowed for this value')
        return total_length * float(number) / 100
    return convert_to_mm(float(number), unit)


def _demo():
    #pattern1 = PatternProtoArea(2.54, obj=THTPad.circle(0, 0, 0.9, 1.8, paste=False))
    #pattern1 = PatternProtoArea(2.54, 2.54, obj=SpikyProto())
    #pattern2 = PatternProtoArea(1.2, 2.0, obj=SMDPad.rect(0, 0, 1.0, 1.8, paste=False))
    #pattern3 = PatternProtoArea(2.54, 1.27, obj=SMDPad.rect(0, 0, 2.3, 1.0, paste=False))
    #pattern3 = EmptyProtoArea(copper_fill=True)
    #stack = TwoSideLayout(pattern2, pattern3)
    #pattern2 = PatternProtoArea(2.54, obj=PoweredProto(), margin=1)
    #pattern3 = PatternProtoArea(2.54, obj=RFGroundProto())
    #stack = PropLayout([pattern2, pattern3], 'h', [0.5, 0.5])
    #pattern = PropLayout([pattern1, stack], 'h', [0.5, 0.5])
    #pattern = PatternProtoArea(2.54, obj=ManhattanPads(2.54))
    #pattern = PatternProtoArea(2.54*1.5, obj=THTFlowerProto())
    #pattern = PatternProtoArea(2.54, obj=THTPad.circle(0, 0, 0.9, 1.8, paste=False))
    #pattern = PatternProtoArea(2.54, obj=PoweredProto())
    pattern = PatternProtoArea(2.54, obj=AlioCell(), margin=2)
    pb = ProtoBoard(50, 47, pattern, mounting_hole_dia=3.2, mounting_hole_offset=5)
    #pb = ProtoBoard(10, 10, pattern1)
    print(pb.pretty_svg())
    pb.layer_stack().save_to_directory('/tmp/testdir')


if __name__ == '__main__':
    _demo()
    #cnt = alphabetic()()
    #for _ in range(32):
    #    for _ in range(26):
    #        print(f'{next(cnt):>2}', end=' ', file=sys.stderr)
    #    print(file=sys.stderr)





from xml.etree import ElementTree
import base64
import json
from pathlib import Path

def _map_primitive(element):
    match element.tag:
        case 'data':
            return base64.b64decode(element.text)
        case 'date':
            return element.text
        case 'true':
            return True
        case 'false':
            return False
        case 'real':
            return float(element.text)
        case 'integer':
            return int(element.text)
        case 'string':
            return element.text
        case 'array':
            return [_map_primitive(child) for child in element]
        case 'dict':
            children = list(element)
            return {k.text: _map_primitive(v) for k, v in zip(children[0::2], children[1::2])}


def parse_shitty_json(data):
    # Parse apple plist XML
    root = ElementTree.fromstring(data)
    return _map_primitive(root[0])


class _SublimeColorschemeSuper:
    def __init__(self, s, by_scope):
        def lookup(default, *scopes):
            for scope in scopes:
                if not (elem := by_scope.get(scope)):
                    continue

                if 'foreground' not in elem:
                    continue

                return elem['foreground']
            return default

        self.background = s.get('background', 'white')
        fg = s.get('foreground', 'black')
        self.bus = lookup(fg, 'constant.other', 'storage.type')
        self.wire = self.lines = lookup(fg, 'constant.other')
        self.no_connect = lookup(fg, 'constant.language', 'variable')
        self.text = lookup(fg, 'constant.numeric', 'constant.numeric.hex', 'storage.type.number')
        self.pin_names = lookup(fg, 'constant.character', 'constant.other')
        self.pin_numbers = fg
        self.values = lookup(fg, 'constant.character.format.placeholder', 'constant.other.placeholder', 'entity.name.tag', 'support.type', 'support.class', 'entity.other.inherited-class')
        self.labels = lookup(fg, 'constant.numeric', 'constant.numeric.hex', 'storage.type.number')
        self.fill = s.get('background')


class TmThemeSchematic(_SublimeColorschemeSuper):
    def __init__(self, data):
        self.theme = parse_shitty_json(data)
        s = self.theme['settings'][0]['settings']
        by_scope = {}
        for elem in self.theme['settings']:
            if 'scope' not in elem:
                continue
            for scope in elem['scope'].split(','):
                by_scope[scope.strip()] = elem.get('settings', {})
        super().__init__(s, by_scope)


class SublimeSchematic(_SublimeColorschemeSuper):
    def __init__(self, data):
        self.theme = json.loads(data)
        s = self.theme['globals']
        by_scope = {}
        for elem in self.theme['rules']:
            for scope in elem['scope'].split(','):
                by_scope[scope.strip()] = elem
        super().__init__(s, by_scope)


if __name__ == '__main__':
    print(parse_shitty_json(Path('/tmp/witchhazelhypercolor.tmTheme').read_text()))





import sys
import math
import warnings
from copy import copy
from itertools import zip_longest, chain
from dataclasses import dataclass, field, replace, KW_ONLY
from collections import defaultdict

from ..utils import LengthUnit, MM, rotate_point, svg_arc, sum_bounds, bbox_intersect, Tag, offset_bounds
from ..layers import LayerStack
from ..graphic_objects import Line, Arc, Flash
from ..apertures import Aperture, CircleAperture, ObroundAperture, RectangleAperture, ExcellonTool
from ..newstroke import Newstroke


class UNDEFINED:
    pass

def sgn(x):
    return -1 if x < 0 else 1


class KeepoutError(ValueError):
    def __init__(self, obj, keepout, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.obj = obj
        self.keepout = keepout


newstroke_font = None


class Board:
    def __init__(self, w=None, h=None, corner_radius=1.5, center=False, default_via_hole=0.4, default_via_diameter=0.8, x=0, y=0, rotation=0, unit=MM):
        self.x, self.y = x, y
        self.rotation = 0
        self.objects = []
        self.outline = []
        self.extra_silk_top = []
        self.extra_silk_bottom = []
        self.keepouts = []
        self.default_via_hole = MM(default_via_hole, unit)
        self.default_via_diameter = MM(default_via_diameter, unit)
        self.unit = unit
        if w or h:
            if w and h:
                self.rounded_rect_outline(w, h, r=corner_radius, center=center)
                self.w, self.h = w, h
            else:
                raise ValueError('Either both, w and h, or neither of them must be given.')
        else:
            self.w = self.h = None

    @property
    def abs_pos(self):
        return self.x, self.y, self.rotation, False

    def add_silk(self, side, obj):
        if side not in ('top', 'bottom'):
            raise ValueError('side must be one of "top" or "bottom".')

        if side == 'top':
            self.extra_silk_top.append(obj)
        else:
            self.extra_silk_bottom.append(obj)

    def add_text(self, *args, **kwargs):
        self.objects.append(Text(*args, **kwargs))

    def add_keepout(self, bbox, unit=MM):
        ((_x_min, _y_min), (_x_max, _y_max)) = bbox
        self.keepouts.append(MM.convert_bounds_from(unit, bbox))

    def add(self, obj, keepout_errors='raise'):
        if keepout_errors not in ('ignore', 'raise', 'warn', 'skip'):
            raise ValueError('keepout_errors must be one of "ignore", "raise", "warn" or "skip".')

        if keepout_errors != 'ignore':
            for ko in self.keepouts:
                if obj.overlaps(ko, unit=MM):
                    if keepout_errors == 'warn':
                        warnings.warn(f'Object with bounds {obj.bounding_box(MM)} [mm] hits one or more keepout areas')
                    elif keepout_errors == 'raise':
                        raise KeepoutError(obj, ko, msg)
                    return

        obj.parent = self
        self.objects.append(obj)

    def via(self, x, y, diameter=None, hole=None, keepout_errors='raise', unit=MM):
        diameter = diameter or unit(self.default_via_dia, MM)
        hole = hole or unit(self.default_via_hole, MM)
        obj = Via(x, y, diameter, hole, unit=unit, keepout_errors=keepout_errors)
        self.add(obj)
        return obj

    def rounded_rect_outline(self, w, h, r=0, x0=None, y0=None, center=False, unit=MM):
        if x0 is None:
            x0 = -w/2 if center else 0
        if y0 is None:
            y0 = -h/2 if center else 0

        ap = CircleAperture(0.05, unit=MM)

        self.outline.append(Line(x0+r, y0, x0+w-r, y0, ap, unit=unit))
        if r:
            self.outline.append(Arc(x0+w-r, y0, x0+w, y0+r, 0, r, False, ap, unit=unit))
        self.outline.append(Line(x0+w, y0+r, x0+w, y0+h-r, ap, unit=unit))
        if r:
            self.outline.append(Arc(x0+w, y0+h-r, x0+w-r, y0+h, -r, 0, False, ap, unit=unit))
        self.outline.append(Line(x0+w-r, y0+h, x0+r, y0+h, ap, unit=unit))
        if r:
            self.outline.append(Arc(x0+r, y0+h, x0, y0+h-r, 0, -r, False, ap, unit=unit))
        self.outline.append(Line(x0, y0+h-r, x0, y0+r, ap, unit=unit))
        if r:
            self.outline.append(Arc(x0, y0+r, x0+r, y0, r, 0, False, ap, unit=unit))

    def layer_stack(self, layer_stack=None):
        if layer_stack is None:
            layer_stack = LayerStack(board_name='proto')

        cache = {}
        for obj in chain(self.objects):
            obj.render(layer_stack, cache)

        layer_stack['mechanical', 'outline'].objects.extend(self.outline)
        layer_stack['top', 'silk'].objects.extend(self.extra_silk_top)
        layer_stack['bottom', 'silk'].objects.extend(self.extra_silk_bottom)

        return layer_stack

    def svg(self, margin=0, arg_unit=MM, svg_unit=MM, force_bounds=None):
        return self.layer_stack().to_svg(margin=margin, arg_unit=arg_unit, svg_unit=svg_unit,
                                                 force_bounds=force_bounds)

    def pretty_svg(self, side='top', margin=0, arg_unit=MM, svg_unit=MM, force_bounds=None, inkscape=False, colors=None):
        return self.layer_stack().to_pretty_svg(side=side, margin=margin, arg_unit=arg_unit, svg_unit=svg_unit,
                                                   force_bounds=force_bounds, inkscape=inkscape, colors=colors)


@dataclass
class Positioned:
    x: float
    y: float
    _: KW_ONLY
    rotation: float = 0.0
    flip: bool = False
    unit: LengthUnit = MM
    parent: object = None

    @property
    def abs_pos(self):
        if self.parent is None:
            px, py, pa, pf = 0, 0, 0, False
        else:
            px, py, pa, pf = self.parent.abs_pos

        return self.x+px, self.y+py, self.rotation+pa, (bool(self.flip) != bool(pf))

    def bounding_box(self, unit=MM):
        stack = LayerStack()
        self.render(stack)
        objects = chain(*(l.objects for l in stack.graphic_layers.values()),
                        stack.drill_pth.objects, stack.drill_npth.objects)
        objects = list(objects)
        #print('foo', type(self).__name__,
        #      [(type(obj).__name__, [prim.bounding_box() for prim in obj.to_primitives(unit)]) for obj in objects], file=sys.stderr)
        return sum_bounds(prim.bounding_box() for obj in objects for prim in obj.to_primitives(unit))

    def overlaps(self, bbox, unit=MM):
        return bbox_intersect(self.bounding_box(unit), bbox)

    @property
    def single_sided(self):
        return True


# The dataclass API is slightly idiotic here, so we have to duplicate the entire thing.
@dataclass(frozen=True)
class FrozenPositioned:
    x: float
    y: float
    _: KW_ONLY
    rotation: float = 0.0
    flip: bool = False
    unit: LengthUnit = MM
    parent: object = None

    @property
    def abs_pos(self):
        if self.parent is None:
            px, py, pa, pf = 0, 0, 0, False
        else:
            px, py, pa, pf = self.parent.abs_pos

        return self.x+px, self.y+py, self.rotation+pa, (bool(self.flip) != bool(pf))

    def bounding_box(self, unit=MM):
        stack = LayerStack()
        self.render(stack)
        objects = chain(*(l.objects for l in stack.graphic_layers.values()),
                        stack.drill_pth.objects, stack.drill_npth.objects)
        objects = list(objects)
        #print('foo', type(self).__name__,
        #      [(type(obj).__name__, [prim.bounding_box() for prim in obj.to_primitives(unit)]) for obj in objects], file=sys.stderr)
        return sum_bounds(prim.bounding_box() for obj in objects for prim in obj.to_primitives(unit))

    def overlaps(self, bbox, unit=MM):
        return bbox_intersect(self.bounding_box(unit), bbox)

    @property
    def single_sided(self):
        return True


@dataclass
class Graphics(Positioned):
    top_copper: list = field(default_factory=list)
    top_mask: list = field(default_factory=list)
    top_silk: list = field(default_factory=list)
    top_paste: list = field(default_factory=list)
    bottom_copper: list = field(default_factory=list)
    bottom_mask: list = field(default_factory=list)
    bottom_silk: list = field(default_factory=list)
    bottom_paste: list = field(default_factory=list)
    drill_npth: list = field(default_factory=list)
    drill_pth: list = field(default_factory=list)

    def render(self, layer_stack, cache=None):
        x, y, rotation, flip = self.abs_pos
        top, bottom = ('bottom', 'top') if flip else ('top', 'bottom')

        for target, source in [
                (layer_stack[top, 'copper'],    self.top_copper),
                (layer_stack[top, 'mask'],      self.top_mask),
                (layer_stack[top, 'silk'],      self.top_silk),
                (layer_stack[top, 'paste'],     self.top_paste),
                (layer_stack[bottom, 'copper'], self.bottom_copper),
                (layer_stack[bottom, 'mask'],   self.bottom_mask),
                (layer_stack[bottom, 'silk'],   self.bottom_silk),
                (layer_stack[bottom, 'paste'],  self.bottom_paste),
                (layer_stack.drill_pth,         self.drill_pth),
                (layer_stack.drill_npth,        self.drill_npth)]:

            for fe in source:
                fe = copy(fe)
                fe.rotate(rotation)
                fe.offset(x, y, self.unit)
                target.objects.append(fe)

    def bounding_box(self, unit=MM):
        if math.isclose(self.rotation, 0, abs_tol=1e-3):
            return offset_bounds(sum_bounds((obj.bounding_box(unit=unit) for obj in chain(
                    self.top_copper,
                    self.top_mask,
                    self.top_silk,
                    self.top_paste,
                    self.bottom_copper,
                    self.bottom_mask,
                    self.bottom_silk,
                    self.bottom_paste,
                    self.drill_npth,
                    self.drill_pth,
                ))), unit(self.x, self.unit), unit(self.y, self.unit))
        else:
            return super().bounding_box(unit)

    @property
    def single_sided(self):
        any_top = self.top_copper or self.top_mask or self.top_paste or self.top_silk
        any_bottom = self.bottom_copper or self.bottom_mask or self.bottom_paste or self.bottom_silk
        any_drill = self.drill_npth or self.drill_pth
        return not (any_drill or (any_top and any_bottom))


@dataclass
class ObjectGroup(Positioned):
    objects: list = field(default_factory=list)

    def render(self, layer_stack, cache=None):
        for obj in self.objects:
            if not isinstance(obj, Positioned):
                raise ValueError(f'ObjectGroup members must be children of Positioned, not {type(obj)}')

            obj.parent = self
            obj.render(layer_stack, cache=cache)

    def bounding_box(self, unit=MM):
        if math.isclose(self.rotation, 0, abs_tol=1e-3):
            return offset_bounds(sum_bounds((obj.bounding_box(unit=unit) for obj in self.objects)),
                                 unit(self.x, self.unit), unit(self.y, self.unit))
        else:
            return super().bounding_box(unit)

    @property
    def single_sided(self):
        return all(obj.single_sided for obj in self.objects)


@dataclass
class Text(Positioned):
    text: str
    font_size: float = 2.5
    stroke_width: float = 0.25
    h_align: str = 'left'
    v_align: str = 'bottom'
    layer: str = 'silk'
    polarity_dark: bool = True

    def render(self, layer_stack, cache=None):
        obj_x, obj_y, rotation, flip = self.abs_pos
        global newstroke_font

        if newstroke_font is None:
            newstroke_font = Newstroke()

        strokes = list(newstroke_font.render(self.text, size=self.font_size))
        if not strokes:
            return

        xs = [x for points in strokes for x, _y in points]
        ys = [y for points in strokes for _x, y in points]
        min_x, min_y, max_x, max_y = min(xs), min(ys), max(xs), max(ys)
        h = self.font_size + self.stroke_width # (max_y - min_y)

        if self.h_align == 'left':
            x0 = 0
        elif self.h_align == 'center':
            x0 = -max_x/2
        elif self.h_align == 'right':
            x0 = -max_x
        else:
            raise ValueError('h_align must be one of "left", "center", or "right".')

        if self.v_align == 'bottom':
            y0 = h
        elif self.v_align == 'middle':
            y0 = h/2
        elif self.v_align == 'top':
            y0 = 0
        else:
            raise ValueError('v_align must be one of "top", "middle", or "bottom".')

        if self.flip:
            x0 += min_x + max_x
            x_sign = -1
        else:
            x_sign = 1

        ap = CircleAperture(self.stroke_width, unit=self.unit)

        for stroke in strokes:
            for (x1, y1), (x2, y2) in zip(stroke[:-1], stroke[1:]):
                obj = Line(x0+x_sign*x1, y0+y1, x0+x_sign*x2, y0+y2, aperture=ap, unit=self.unit, polarity_dark=self.polarity_dark)
                obj.rotate(rotation)
                obj.offset(obj_x, obj_y)
                layer_stack['bottom' if flip else 'top', self.layer].objects.append(obj)

    def bounding_box(self, unit=MM):
        approx_w = len(self.text)*self.font_size*0.75 + self.stroke_width
        approx_h = self.font_size + self.stroke_width

        if self.h_align == 'left':
            x0 = 0
        elif self.h_align == 'center':
            x0 = -approx_w/2
        elif self.h_align == 'right':
            x0 = -approx_w

        if self.v_align == 'top':
            y0 = 0
        elif self.v_align == 'middle':
            y0 = -approx_h/2
        elif self.v_align == 'bottom':
            y0 = -approx_h

        return (self.x+x0, self.y+y0), (self.x+x0+approx_w, self.y+y0+approx_h)


@dataclass(frozen=True, slots=True)
class PadStackAperture:
    aperture: Aperture
    side: str
    layer: str
    offset_x: float = 0 # in PadStack units
    offset_y: float = 0
    rotation: float = 0
    invert: bool = False


@dataclass(frozen=True, slots=True)
class PadStack:
    _: KW_ONLY
    unit: LengthUnit = MM

    @property
    def apertures(self):
        raise NotImplementedError()

    def flashes(self, x, y, rotation: float = 0, flip: bool = False):
        for ap in self.apertures:
            aperture = ap.aperture.rotated(ap.rotation + rotation)
            fl = Flash(ap.offset_x, ap.offset_y, aperture, polarity_dark=not ap.invert, unit=self.unit)
            fl.rotate(rotation)
            fl.offset(x, y)
            side = ap.side
            if flip:
                side = {'top': 'bottom', 'bottom': 'top'}.get(side, side)
            yield side, ap.layer, fl

    def render(self, layer_stack, x, y, rotation: float = 0, flip: bool = False):
        for side, layer, flash in self.flashes(x, y, rotation, flip):
            if side == 'drill' and layer == 'plated':
                layer_stack.drill_pth.objects.append(flash)

            elif side == 'drill' and layer == 'nonplated':
                layer_stack.drill_npth.objects.append(flash)

            elif (side, layer) in layer_stack:
                layer_stack[side, layer].objects.append(flash)

    @property
    def single_sided(self):
        return len({ap.side for ap in self.apertures}) <= 1


@dataclass(frozen=True, slots=True)
class SMDStack(PadStack):
    aperture: Aperture
    mask_expansion: float = 0.0
    paste_expansion: float = 0.0
    paste: bool = True
    flip: bool = False

    @property
    def side(self):
        return 'bottom' if self.flip else 'top'

    @property
    def apertures(self):
        yield PadStackAperture(self.aperture, self.side, 'copper')
        yield PadStackAperture(self.aperture.dilated(self.mask_expansion, self.unit), self.side, 'mask')
        if self.paste:
            yield PadStackAperture(self.aperture.dilated(self.paste_expansion, self.unit), self.side, 'paste')

    @classmethod
    def rect(kls, w, h, rotation=0, mask_expansion=0.0, paste_expansion=0.0, paste=True, flip=False, unit=MM):
        ap = RectangleAperture(w, h, unit=unit).rotated(rotation)
        return kls(ap, mask_expansion, paste_expansion, paste, flip, unit=unit)

    @classmethod
    def circle(kls, dia, mask_expansion=0.0, paste_expansion=0.0, paste=True, flip=False, unit=MM):
        return kls(CircleAperture(dia, unit=unit), mask_expansion, paste_expansion, paste, flip, unit=unit)
    

@dataclass(frozen=True, slots=True)
class MechanicalHoleStack(PadStack):
    drill_dia: float
    mask_expansion: float = 0.0
    mask_aperture = None

    @property
    def apertures(self):
        mask_aperture = self.mask_aperture or CircleAperture(self.drill_dia + self.mask_expansion, unit=self.unit)
        yield PadStackAperture(mask_aperture, 'top', 'mask')
        yield PadStackAperture(mask_aperture, 'bottom', 'mask')

    @property
    def single_sided(self):
        return False


@dataclass(frozen=True, slots=True)
class THTPad(PadStack):
    drill_dia: float
    pad_top: SMDStack
    pad_bottom: SMDStack = None
    aperture_inner: Aperture = UNDEFINED
    plated: bool = True

    def __post_init__(self):
        if self.pad_bottom is None:
            object.__setattr__(self, 'pad_bottom', replace(self.pad_top, flip=True))
        
        if self.aperture_inner is UNDEFINED:
            object.__setattr__(self, 'aperture_inner', self.pad_top.aperture)

        if self.pad_top.flip:
            raise ValueError('top pad cannot be flipped')

    @property
    def plating(self):
        return 'plated' if self.plated else 'nonplated'

    @property
    def apertures(self):
        yield from self.pad_top.apertures
        yield from self.pad_bottom.apertures
        if self.aperture_inner is not None:
            yield PadStackAperture(self.aperture_inner, 'inner', 'copper')
        yield PadStackAperture(ExcellonTool(self.drill_dia, plated=self.plated, unit=self.unit), 'drill', self.plating)

    @property
    def single_sided(self):
        return False

    @classmethod
    def rect(kls, drill_dia, w, h, rotation=0, mask_expansion=0.0, paste_expansion=0.0, paste=True, plated=True, unit=MM):
        pad = SMDStack.rect(w, h, rotation, mask_expansion, paste_expansion, paste, unit=unit)
        return kls(drill_dia, pad, plated=plated)

    @classmethod
    def circle(kls, drill_dia, dia, rotation=0, mask_expansion=0.0, paste_expansion=0.0, paste=True, plated=True, unit=MM):
        pad = SMDStack.circle(dia, mask_expansion, paste_expansion, paste, unit=unit)
        return kls(drill_dia, pad, plated=plated)

    @classmethod
    def obround(kls, drill_dia, w, h, rotation=0, mask_expansion=0.0, paste_expansion=0.0, paste=True, plated=True, unit=MM):
        ap = ObroundAperture(w, h, unit=unit).rotated(rotation)
        pad = SMDStack(ap, mask_expansion, paste_expansion, paste, unit=unit)
        return kls(drill_dia, pad, plated=plated)

@dataclass(frozen=True, slots=True)
class ThroughViaStack(PadStack):
    hole: float
    dia: float = None
    tented: bool = True

    def __post_init__(self):
        if self.dia == None:
            object.__setattr__(self, 'dia', self.hole*2)

    @property
    def single_sided(self):
        return False

    @property
    def apertures(self):
        copper_aperture = CircleAperture(self.dia, unit=self.unit)
        yield PadStackAperture(copper_aperture, 'top', 'copper')
        yield PadStackAperture(copper_aperture, 'bottom', 'copper')
        yield PadStackAperture(copper_aperture, 'inner', 'copper')
        if self.tented:
            yield PadStackAperture(copper_aperture, 'top', 'mask')
            yield PadStackAperture(copper_aperture, 'bottom', 'mask')
        yield PadStackAperture(ExcellonTool(self.hole, plated=True, unit=self.unit), 'drill', 'plated')


@dataclass(frozen=True, slots=True)
class Via(FrozenPositioned):
    pad_stack: PadStack

    def render(self, layer_stack, cache=None):
        x, y, rotation, flip = self.abs_pos
        self.pad_stack.render(layer_stack, x, y, rotation, flip)

    @classmethod
    def at(kls, x, y, hole, dia=None, tented=True, unit=MM):
        return kls(x, y, ThroughViaStack(hole, dia, tented, unit=unit), unit=unit)


@dataclass
class Pad(Positioned):
    pad_stack: PadStack

    def render(self, layer_stack, cache=None):
        x, y, rotation, flip = self.abs_pos
        self.pad_stack.render(layer_stack, x, y, rotation, flip)

    @property
    def single_sided(self):
        return self.pad_stack.single_sided


@dataclass
class Trace:
    width: float
    start: object = None
    end: object = None
    waypoints: [(float, float)] = field(default_factory=list)
    style: str = 'oblique'
    orientation: [str] = tuple() # 'cw' or 'ccw'
    roundover: float = 0
    side: str = 'top'
    unit: LengthUnit = MM
    parent: object = None

    DIRECT = 'direct'
    OBLIQUE = 'oblique'
    ORTHO = 'ortho'

    CW = 'cw'
    CCW = 'ccw'

    def _route(self, p1, p2, orientation):
        x1, y1 = p1
        x2, y2 = p2
        dx = x2-x1
        dy = y2-y1

        yield p1

        if self.style == 'direct' or \
                math.isclose(x1, x2, abs_tol=1e-6) or math.isclose(y1, y2, abs_tol=1e-6) or \
                (self.style == 'oblique' and math.isclose(dx, dy, abs_tol=1e-6)):
            return

        p = (abs(dy) > abs(dx)) == ((dx >= 0) == (dy >= 0))
        if self.style == 'oblique':
            if p == (orientation == 'cw'):
                if abs(dy) > abs(dx):
                    yield (x1, y1+sgn(dy)*(abs(dy)-abs(dx)))
                else:
                    yield (x1+sgn(dx)*(abs(dx)-abs(dy)), y1)
            else:
                if abs(dy) > abs(dx):
                    yield (x2, y1+sgn(dy)*abs(dx))
                else:
                    yield (x1+sgn(dx)*abs(dy), y2)

        else: # self.style == 'ortho'
            if p == (orientation == 'cw'):
                if abs(dy) > abs(dx):
                    yield (x1, y2)
                else:
                    yield (x2, y1)
            else:
                if abs(dy) > abs(dx):
                    yield (x2, y1)
                else:
                    yield (x1, y2)

    @classmethod
    def _midpoint(kls, p1, p2):
        x1, y1 = p1
        x2, y2 = p2
        dx = x2 - x1
        dy = y2 - y1
        xm = x1 + dx / 2
        ym = y1 + dy / 2
        return (xm, ym)

    @classmethod
    def _point_on_line(kls, p1, p2, dist_from_p1):
        x1, y1 = p1
        x2, y2 = p2
        dx = x2 - x1
        dy = y2 - y1
        dist = math.dist(p1, p2)
        if math.isclose(dist, 0, abs_tol=1e-6):
            return p2
        xm = x1 + dx / dist * dist_from_p1
        ym = y1 + dy / dist * dist_from_p1
        return (xm, ym)

    @classmethod
    def _angle_between(kls, p1, p2, p3):
        x1, y1 = p1
        x2, y2 = p2
        x3, y3 = p3
        x1, y1 = x1 - x2, y1 - y2
        x3, y3 = x3 - x2, y3 - y2
        dot_product = x1*x3 + y1*y3
        l1 = math.hypot(x1, y1)
        l2 = math.hypot(x3, y3)
        norm = dot_product / l1 / l2
        return math.acos(min(1, max(-1, norm)))

    def _round_over(self, points, aperture):
        if math.isclose(self.roundover, 0, abs_tol=1e-6) or len(points) <= 2:
            import sys
            for p1, p2 in zip(points[:-1], points[1:]):
                yield Line(*p1, *p2, aperture=aperture, unit=self.unit)
            return
        # here: len(points) >= 3

        line_b = Line(*points[0], *self._midpoint(points[0], points[1]), aperture=aperture, unit=self.unit)

        for p1, p2, p3 in zip(points[:-2], points[1:-1], points[2:]):
            x1, y1 = p1
            x2, y2 = p2
            x3, y3 = p3
            xa, ya = pa = self._midpoint(p1, p2)
            xb, yb = pb = self._midpoint(p2, p3)
            la = math.dist(pa, p2)
            lb = math.dist(p2, pb)

            alpha = self._angle_between(p1, p2, p3)
            if alpha == 0:
                l = Line(line_b.x1, line_b.y1, *p2, aperture=aperture, unit=self.unit)
                line_b = Line(*p2, *pb, aperture=aperture, unit=self.unit)
                yield l
                continue
            tr = self.roundover/math.tan(alpha/2)
            t = min(la, lb, tr)
            r = t*math.tan(alpha/2)

            xs, ys = ps = self._point_on_line(p2, pa, t)
            xe, ye = pe = self._point_on_line(p2, pb, t)

            if math.isclose(t, la, abs_tol=1e-6):
                if not math.isclose(line_b.curve_length(), 0, abs_tol=1e-6):
                    yield line_b
                xs, ys = ps = pa
            else:
                yield Line(line_b.x1, line_b.y1, xs, ys, aperture=aperture, unit=self.unit)

            if math.isclose(t, lb, abs_tol=1e-6):
                xe, ye = pe = pb
            line_b = Line(*pe, *pb, aperture=aperture, unit=self.unit)

            if math.isclose(r, 0, abs_tol=1e-6):
                continue

            xc = -(y2 - ys) / t * r
            yc = +(x2 - xs) / t * r

            xsr = xs - x2
            ysr = ys - y2
            xer = xe - x2
            yer = ye - y2
            cross_product_z = xsr * yer - ysr * xer

            clockwise = cross_product_z > 0
            if clockwise:
                xc, yc = -xc, -yc
            
            yield Arc(*ps, *pe, xc, yc, clockwise, aperture=aperture, unit=self.unit)

        yield Line(line_b.x1, line_b.y1, x3, y3, aperture=aperture, unit=self.unit)
        
    def to_graphic_objects(self):
        start, end = self.start, self.end

        if not isinstance(start, tuple):
            *start, _rotation, _flip = start.abs_pos
        if not isinstance(end, tuple):
            *end, _rotation, _flip = end.abs_pos

        aperture = CircleAperture(diameter=self.width, unit=self.unit)

        points_in = [start, *self.waypoints, end]

        points = []
        for p1, p2, orientation in zip_longest(points_in[:-1], points_in[1:], self.orientation):
            points.extend(self._route(p1, p2, orientation))
        points.append(p2)

        return self._round_over(points, aperture)

    def render(self, layer_stack, cache=None):
        layer_stack[self.side, 'copper'].objects.extend(self.to_graphic_objects())

def _route_demo():
    from ..utils import setup_svg, Tag

    def pd_obj(objs):
        objs = list(objs)
        yield f'M {objs[0].x1}, {objs[0].y1}'
        for obj in objs:
            if isinstance(obj, Line):
                yield f'L {obj.x2}, {obj.y2}'
            else:
                assert isinstance(obj, Arc)
                yield svg_arc(obj.p1, obj.p2, obj.center_relative, obj.clockwise)

    pd = lambda points: f'M {points[0][0]}, {points[0][1]} ' + ' '.join(f'L {x}, {y}' for x, y in points[1:])

    font = Newstroke()

    tags = []
    for n in range(0, 8*6):
        theta = 2*math.pi / (8*6) * n
        dx, dy = math.cos(theta), math.sin(theta)

        strokes = list(font.render(f'α={n/(8*6)*360}', size=0.2))
        xs = [x for st in strokes for x, _y in st]
        ys = [y for st in strokes for _x, y in st]
        min_x, min_y, max_x, max_y = min(xs), min(ys), max(xs), max(ys)

        xf = f'translate({n//6*1.1 + 0.1} {n%6*1.3 + 0.3}) scale(0.5 0.5) translate(1 1)'
        txf = f'{xf} translate(0 -1.2) translate({-(max_x-min_x)/2} {-max_y})'

        tags.append(Tag('circle', cx='0', cy='0', r='1',
                        fill='none', stroke='black', opacity='0.5', stroke_width='0.01',
                        transform=xf))
        tags.append(Tag('path',
                        fill='none',
                        stroke='black', opacity='0.5', stroke_width='0.02', stroke_linejoin='round', stroke_linecap='round',
                        transform=txf, d=' '.join(pd(points) for points in strokes)))

        #for r in [0.0, 0.1, 0.2, 0.3]:
        for r in [0, 0.2]:
            #tr = Trace(0.1, style='ortho', roundover=r, start=(0, 0), end=(dx, dy))
            tr = Trace(0.1, style='oblique', roundover=r, start=(dx, dy), end=(0, 0))
            #points_cw = list(tr._route((0, 0), (dx, dy), 'cw')) + [(dx, dy)]
            #points_ccw = list(tr._route((0, 0), (dx, dy), 'ccw')) + [(dx, dy)]
            tr.orientation = ['cw']
            objs_cw = tr._to_graphic_objects()
            tr.orientation = ['ccw']
            objs_ccw = tr._to_graphic_objects()

            tags.append(Tag('path',
                            fill='none',
                            stroke='red', stroke_width='0.01', stroke_linecap='round',
                            transform=xf, d=' '.join(pd_obj(objs_cw))))
            tags.append(Tag('path',
                            fill='none',
                            stroke='blue', stroke_width='0.01', stroke_linecap='round',
                            transform=xf, d=' '.join(pd_obj(objs_ccw))))
            #tags.append(Tag('path',
            #                fill='none',
            #                stroke='red', stroke_width='0.01', stroke_linecap='round',
            #                transform=xf, d=pd(points_cw)))
            #tags.append(Tag('path',
            #                fill='none',
            #                stroke='blue', stroke_width='0.01', stroke_linecap='round',
            #                transform=xf, d=pd(points_ccw)))


    print(setup_svg([Tag('g', tags, transform='scale(20 20)')], [(0, 0), (20*10*1.1 + 0.1, 20*10*1.3 + 0.1)]))


def _board_demo():
    b = Board(100, 80)
    p1 = THTPad.rect(10, 10, 0.9, 1.8)
    b.add(p1)
    p2 = THTPad.rect(20, 15, 0.9, 1.8)
    b.add(p2)
    b.add(Trace(0.5, p1, p2, style='ortho', roundover=1.5))
    b.add_text(50, 50, 'Foobar')
    print(b.pretty_svg())
    b.layer_stack().save_to_directory('/tmp/testdir')


if __name__ == '__main__':
    _board_demo()
    #_route_demo()





import textwrap

import copy
from dataclasses import MISSING, replace, fields
from .sexp import *


SEXP_END = type('SEXP_END', (), {})


class AtomChoice:
    def __init__(self, *choices):
        self.choices = choices

    def __contains__(self, value):
        return value in self.choices

    def __atoms__(self):
        return self.choices

    def __map__(self, obj, parent=None):
        obj, = obj
        if obj not in self:
            raise TypeError(f'Invalid atom {obj} for {type(self)}, valid choices are: {", ".join(map(str, self.choices))}')
        return obj

    def __sexp__(self, value):
        yield value

    def __str__(self):
        choices = '|'.join(map(str, self.choices))
        return f'AtomChoice({choices})'


class Flag:
    def __init__(self, atom=None, invert=None):
        self.atom, self.invert = atom, invert

    def __bind_field__(self, field):
        if self.atom is None:
            self.atom = Atom(field.name)
        if self.invert is None:
            self.invert = bool(field.default)

    def __atoms__(self):
        return [self.atom]

    def __map__(self, obj, parent=None):
        return not self.invert

    def __sexp__(self, value):
        if bool(value) == (not self.invert):
            yield self.atom
    
    def __str__(self):
        if self.invert is not None:
            return f'Flag({self.atom}/{self.invert})'
        return f'Flag({self.atom})'


def sexp(t, v):
    try:
        if v is None:
            return []
        elif t in (int, float, str, Atom):
            return [t(v)]
        elif hasattr(t, '__sexp__'):
            return list(t.__sexp__(v))
        elif isinstance(t, list):
            t, = t
            return [sexp(t, elem) for elem in v]
        else:
            raise TypeError(f'Python type {t} of value {v!r} has no defined s-expression serialization')

    except MappingError as e:
        raise e

    except Exception as e:
        raise MappingError(f'Error trying to serialize {textwrap.shorten(str(v), width=120)} into type {t}', t, v) from e


class MappingError(TypeError):
    def __init__(self, msg, t, sexp):
        super().__init__(msg)
        self.t, self.sexp = t, sexp

def map_sexp(t, v, parent=None, path=''):
    try:
        if t is not Atom and hasattr(t, '__map__'):
            return t.__map__(v, parent=parent)

        elif t in (int, float, str, Atom):
            v, = v
            if not isinstance(v, t):
                types = set({type(v), t})
                if types == {int, float} or types == {str, Atom}:
                    v = t(v)
                else:
                    raise TypeError(f'Cannot map s-expression value {v} of type {type(v)} to Python type {t}')
            return v

        elif isinstance(t, list):
            t, = t
            return [map_sexp(t, elem, parent=parent, path=f'{path}/{t}') for elem in v]

        else:
            raise TypeError(f'Python type {t} has no defined s-expression deserialization')

    except MappingError as e:
        raise e

    except Exception as e:
        raise MappingError(f'Error at {path} trying to map {textwrap.shorten(str(v), width=60)} into type {t}', t, v) from e


class WrapperType:
    def __init__(self, next_type):
        self.next_type = next_type

    def __bind_field__(self, field):
        self.field = field
        if self.next_type is not Atom:
            getattr(self.next_type, '__bind_field__', lambda x: None)(field)

    def __atoms__(self):
        if hasattr(self, 'name_atom'):
            return [self.name_atom]
        elif self.next_type is Atom:
            return []
        else:
            return getattr(self.next_type, '__atoms__', lambda: [])()

class Named(WrapperType):
    def __init__(self, next_type, name=None, omit_empty=True):
        super().__init__(next_type)
        self.name_atom = Atom(name) if name else None
        self.omit_empty = omit_empty

    def __bind_field__(self, field):
        if self.next_type is not Atom:
            getattr(self.next_type, '__bind_field__', lambda x: None)(field)
        if self.name_atom is None:
            self.name_atom = Atom(field.name)

    def __map__(self, obj, parent=None, path=''):
        k, *obj = obj
        if self.next_type in (int, float, str, Atom) or isinstance(self.next_type, AtomChoice):
            return map_sexp(self.next_type, [*obj], parent=parent, path=f'{path}/{self.name_atom}')
        else:
            return map_sexp(self.next_type, obj, parent=parent, path=f'{path}/{self.name_atom}')

    def __sexp__(self, value):
        value = sexp(self.next_type, value)
        if value is None:
            return

        if self.omit_empty and not value:
            return

        yield [self.name_atom, *value]

    def __str__(self):
        return f'Named={self.name_atom}({self.next_type})'


class Rename(WrapperType):
    def __init__(self, next_type, name=None):
        super().__init__(next_type)
        self.name_atom = Atom(name) if name else None

    def __bind_field__(self, field):
        if self.name_atom is None:
            self.name_atom = Atom(field.name)
        if hasattr(self.next_type, '__bind_field__'):
            self.next_type.__bind_field__(field)

    def __map__(self, obj, parent=None, path=''):
        return map_sexp(self.next_type, obj, parent=parent, path=f'{path}/{self.name_atom}')

    def __sexp__(self, value):
        value, = sexp(self.next_type, value)
        if self.next_type in (str, float, int, Atom): 
            yield [self.name_atom, *value]
        else:
            key, *rest = value
            yield [self.name_atom, *rest]

    def __str__(self):
        return f'Rename={self.name_atom}({self.next_type})'


class OmitDefault(WrapperType):
    def __bind_field__(self, field):
        getattr(self.next_type, '__bind_field__', lambda x: None)(field)
        if field.default_factory != MISSING:
            self.default = field.default_factory()
        else:
            self.default = field.default

    def __map__(self, obj, parent=None, path=''):
        return map_sexp(self.next_type, obj, parent=parent, path=path)

    def __sexp__(self, value):
        if value != self.default:
            yield from sexp(self.next_type, value)

    def __str__(self):
        return f'OmitDefault({self.field})'


class YesNoAtom:
    def __init__(self, yes=Atom.yes, no=Atom.no):
        self.yes, self.no = yes, no

    def __map__(self, value, parent=None):
        if not value: # compatibility with legacy flag style
            return False
        value, = value
        return value == self.yes

    def __sexp__(self, value):
        yield self.yes if value else self.no


class LegacyCompatibleFlag:
    '''Variant of YesNoAtom that accepts both the `(flag <yes/no>)` variant and the bare `flag` variant for compatibility.'''

    def __init__(self, yes=Atom.yes, no=Atom.no, value_when_empty=True):
        self.yes, self.no = yes, no
        self.value_when_empty = value_when_empty

    def __map__(self, value, parent=None):
        if value == []:
            return self.value_when_empty

        value, = value
        return value == self.yes

    def __sexp__(self, value):
        yield self.yes if value else self.no


class Wrap(WrapperType):
    def __map__(self, value, parent=None, path=''):
        value, = value
        return map_sexp(self.next_type, value, parent=parent, path=path)

    def __sexp__(self, value):
        for inner in sexp(self.next_type, value):
            yield [inner]

    def __str__(self):
        return f'Wrap({self.next_type})'


class Array(WrapperType):
    def __map__(self, value, parent=None, path=''):
        return [map_sexp(self.next_type, [elem], parent=parent, path=path) for elem in value]
    
    def __sexp__(self, value):
        for e in value:
            yield from sexp(self.next_type, e)

    def __str__(self):
        return f'Array({self.next_type})'


class Untagged(WrapperType):
    def __map__(self, value, parent=None, path=''):
        value, = value
        return self.next_type.__map__([self.next_type.name_atom, *value], parent=parent, path=path)
    
    def __sexp__(self, value):
        for inner in sexp(self.next_type, value):
            _tag, *rest = inner
            yield rest

    def __str__(self):
        return f'Untagged({self.next_type})'

class List(WrapperType):
    def __bind_field__(self, field):
        self.attr = field.name

    def __map__(self, value, parent, path=''):
        l = getattr(parent, self.attr, [])
        mapped = map_sexp(self.next_type, value, parent=parent, path=f'{path}/{self.attr}')
        l.append(mapped)
        setattr(parent, self.attr, l)

    def __sexp__(self, value):
        for elem in value:
            yield from sexp(self.next_type, elem)

    def __str__(self):
        return f'List@{self.attr}({self.next_type})'


class _SexpTemplate:
    @staticmethod
    def __atoms__(kls):
        return [kls.name_atom]

    @staticmethod
    def __map__(kls, value, *args, parent=None, path='', **kwargs):
        positional = iter(kls.positional)
        inst = kls(*args, **kwargs)

        for v in value[1:]: # skip key
            if isinstance(v, Atom) and v in kls.keys:
                name, etype = kls.keys[v]
                mapped = map_sexp(etype, [v], parent=inst, path=f'{path}/{kls.name_atom}')
                if mapped is not None:
                    setattr(inst, name, mapped)

            elif isinstance(v, list):
                key = v[0]
                if key in kls.keys:
                    name, etype = kls.keys[key]
                    mapped = map_sexp(etype, v, parent=inst, path=f'{path}/{kls.name_atom}')
                    if mapped is not None:
                        setattr(inst, name, mapped)

                elif hasattr(inst, '__catchall__'):
                    inst.__catchall__(v, path=f'{path}/{kls.name_atom}')

                else:
                    #print('class has keys:')
                    #print('\n'.join(map(str, kls.keys)))
                    raise TypeError(f'Unhandled keyed argument {v!r} while parsing {kls}')

            else:
                try:
                    pos_key = next(positional)
                    setattr(inst, pos_key.name, v)
                except StopIteration:
                    raise TypeError(f'Unhandled positional argument {v!r} while parsing {kls}')

        getattr(inst, '__after_parse__', lambda x: None)(parent)
        return inst

    @staticmethod
    def __sexp__(kls, value):
        getattr(value, '__before_sexp__', lambda: None)()

        out = [kls.name_atom]
        for f in fields(kls):
            if f.type is SEXP_END:
                break
            out += sexp(f.type, getattr(value, f.name))
        yield out

    @staticmethod
    def parse(kls, data, *args, **kwargs):
        return kls.__map__(parse_sexp(data), *args, **kwargs)

    @staticmethod
    def sexp(self):
        return next(self.__sexp__(self))

    @staticmethod
    def __deepcopy__(self, memo):
        return replace(self, **{f.name: copy.deepcopy(getattr(self, f.name), memo) for f in fields(self) if not f.kw_only})

    @staticmethod
    def __copy__(self):
        # Even during a shallow copy, we need to deep copy any fields whose types have a __before_sexp__ method to avoid
        # those from being called more than once on the same object.
        return replace(self, **{f.name: copy.copy(getattr(self, f.name)) for f in fields(self) if not f.kw_only and hasattr(f.type, '__before_sexp__')})

def sexp_type(name=None):
    def register(cls):
        cls = dataclass(cls)
        cls.name_atom = Atom(name) if name is not None else None
        for key in '__sexp__', '__map__', '__atoms__', 'parse':
            if not hasattr(cls, key):
                setattr(cls, key, classmethod(getattr(_SexpTemplate, key)))

        for key in 'sexp', '__deepcopy__', '__copy__':
            if not hasattr(cls, key):
                setattr(cls, key, getattr(_SexpTemplate, key))

        cls.positional = []
        cls.keys = {}
        for f in fields(cls):
            f_type = f.type
            if f_type is SEXP_END:
                break

            if hasattr(f_type, '__bind_field__'):
                f_type.__bind_field__(f)

            atoms = getattr(f_type, '__atoms__', lambda: [])
            atoms = list(atoms())
            for atom in atoms:
                cls.keys[atom] = (f.name, f_type)
            if not atoms:
                cls.positional.append(f)

        return cls
    return register





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Jan Sebastian Götte <code@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Gerbonara
=========

gerbonara provides utilities for working with PCB artwork files in Gerber/RS274-X, XNC/Excellon and IPC-356 formats. It
includes convenience functions to match file names to layer types that match the default settings of a number of common
EDA tools.
"""

from .rs274x import GerberFile
from .excellon import ExcellonFile
from .ipc356 import Netlist
from .layers import LayerStack
from .utils import MM, Inch
from importlib.metadata import version

__version__ = version('gerbonara')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# copyright 2014 Hamilton Kibbe <ham@hamiltonkib.be>
# Modified from parser.py by Paulo Henrique Silva <ph.silva@gmail.com>
# Copyright 2022 Jan Sebastian Götte <gerbonara@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from dataclasses import dataclass
import math
import re
from enum import Enum
import warnings
from dataclasses import dataclass
from pathlib import Path

from .cam import CamFile, FileSettings
from .utils import MM, Inch, LengthUnit, rotate_point


class Netlist(CamFile):
    def __init__(self, test_records=None, conductors=None, outlines=None, comments=None, adjacency=None,
            params=None, import_settings=None, original_path=None, generator_hints=None):
        super().__init__(original_path=original_path, layer_name='netlist', import_settings=import_settings)
        self.test_records = test_records or []
        self.conductors = conductors or []
        self.outlines = outlines or []
        self.comments = comments or []
        self.adjacency = adjacency or {}
        self.params = params or {}
        self.generator_hints = generator_hints or []

    def merge(self, other, our_prefix=None, their_prefix=None):
        ''' Merge other netlist into this netlist. The respective net names are prefixed with the given prefixes
        (default: None). Garbles other. '''
        if other is None:
            return

        if not isinstance(other, Netlist):
            raise TypeError(f'Can only merge Netlist with other Netlist, not {type(other)}')

        self.prefix_nets(our_prefix)
        other.prefix_nets(our_prefix)

        self.test_records.extend(other.test_records)
        self.conductors.extend(other.conductors)
        self.outlines.extend(other.outlines)
        self.comments.extend(other.comments)
        self.adjacency.update(other.adjacency)
        self.params.update(other.params)

        self.params['JOB'] = 'Gerbonara IPC-356 merge'
        self.params['TITLE'] = 'Gerbonara IPC-356 merge'

        for key in 'CODE', 'NUM', 'REV', 'VER':
            if key in self.params:
                del self.params[key]

    def prefix_nets(self, prefix):
        if not prefix:
            return

        for record in self.test_records:
            if record.net_name:
                record.net_name = prefix + record.net_name

        for conductor in self.conductors:
            if conductor.net_name:
                conductor.net_name = prefix + conductor.net_name

        new_adjacency = {}
        for key in self.adjacency:
            new_adjacency[prefix + key] = [ prefix + name for name in self.adjacency[key] ]
        self.adjacency = new_adjacency

    def offset(self, dx=0,  dy=0, unit=MM):
        for obj in self.objects:
            obj.offset(dx, dy, unit)

    def rotate(self, angle:'radian', center=(0,0), unit=MM):
        cx, cy = center

        for obj in self.objects:
            obj.rotate(angle, cx, cy, unit)

    def __str__(self):
        name = f'{self.original_path.name} ' if self.original_path else ''
        return f'<IPC-356 Netlist {name}with {len(self.test_records)} records, {len(self.conductors)} conductors and {len(self.outlines)} outlines>'

    def __repr__(self):
        return str(self)

    @property
    def objects(self):
        yield from self.test_records
        yield from self.conductors
        yield from self.outlines

    @classmethod
    def open(kls, filename):
        path = Path(filename)
        parser = NetlistParser()
        return parser.parse(path.read_text(), path)

    @classmethod
    def from_string(kls, data, filename=None):
        parser = NetlistParser()
        return parser.parse(data, Path(filename))

    def save(self, filename, settings=None, drop_comments=True):
        with open(filename, 'wb') as f:
            f.write(self.write_to_bytes(settings, drop_comments=drop_comments))

    def write_to_bytes(self, settings=None, drop_comments=True, job_name=None):
        if settings is None:
            settings = self.import_settings.copy() or FileSettings()
            settings.zeros = None
            settings.number_format = (5,6)
        return '\n'.join(self._generate_lines(settings, drop_comments=drop_comments)).encode('utf-8')

    def _generate_lines(self, settings, drop_comments, job_name=None):
        yield 'C  IPC-D-356 generated by Gerbonara'
        yield 'C'
        yield f'P  JOB {self.params.get("JOB", "Gerbonara netlist export")}'
        yield 'P  UNITS CUST 0' if settings.unit == Inch else 'P  UNITS CUST 1' 

        if not drop_comments:
            for comment in self.comments:
                yield f'C  {comment}'

        for name, value in self.params.items():
            if name == 'JOB':
                continue

            yield f'P  {name} {value!s}'

        net_name_map = {
                name: f'NNAME{i}' for i, name in enumerate(
                    name for name in self.net_names() if len(name) > 14
                    ) }

        yield 'C'
        yield 'C  Net name mapping:'
        yield 'C'
        for name, alias in net_name_map.items():
            yield f'P  {alias} {name}'

        yield 'C'
        yield 'C  Test records:'
        yield 'C'

        for record in self.test_records:
            yield from record.format(settings, net_name_map)

        if self.conductors:
            yield 'C'
            yield 'C  Conductors:'
            yield 'C'
            for conductor in self.conductors:
                yield from conductor.format(settings, net_name_map)

        if self.outlines:
            yield 'C'
            yield 'C  Outlines:'
            yield 'C'
            for outline in self.outlines:
                yield from outline.format(settings)

        if self.adjacency:
            yield 'C'
            yield 'C  Adjacency data:'
            yield 'C'
            done = set()
            for net, others in self.adjacency.items():
                others_filtered = [ other for other in others if (net, other) not in done and (other, net) not in done ]

                line = '379'
                for net in self.nets:
                    if len(line) + 1 + len(net) > 80:
                        yield line
                        line = f'079 {net}'
                    else:
                        line += f' {net}'
                yield line

    def net_names(self):
        nets = { record.net_name for record in self.test_records }
        nets -= {None}
        return nets

    def vias(self):
        for record in self.test_records:
            if record.is_via:
                yield record

    def reference_designators(self):
        names = { record.ref_des for record in self.test_records }
        names -= {None}
        return names

    def records_by_reference(self, reference_designator):
        for record in self.test_records:
            if record.ref_des == reference_designator:
                yield record

    def records_by_net_name(self, net_name):
        for record in self.test_records:
            if record.net_name == net_name:
                yield record

    def conductors_by_net_name(self, net_name):
        for conductor in self.conductos:
            if conductor.net_name == net_name:
                yield conductor

    def conductors_by_layer(self, layer : int):
        for conductor in self.conductos:
            if conductor.layer == layer:
                yield conductor


class NetlistParser(object):
    # Good resources on IPC-356 syntax are:
    # https://www.downstreamtech.com/downloads/IPCD356_Simplified.pdf
    # https://web.pa.msu.edu/hep/atlas/l1calo/hub/hardware/components/circuit_board/ipc_356a_net_list.pdf

    def __init__(self):
        self.has_unit = False
        self.settings = FileSettings()
        self.net_names = {}
        self.params = {}
        self.comments = []
        self.test_records = []
        self.conductors = []
        self.adjacency = {}
        self.outlines = []
        self.eof = False
        self.generator_hints = []

    def warn(self, msg, kls=SyntaxWarning):
        warnings.warn(f'{self.filename}:{self.start_line}: {msg}', kls)

    def assert_unit(self):
        if not self.has_unit:
            raise SyntaxError('IPC-356 netlist file does not contain unit specification before first entry')

    def parse(self, data, path=None):
        self.filename = path.name

        try:
            oldline = ''
            for lineno, line in enumerate(data.splitlines()):
                # Check for existing multiline data...
                if oldline:
                    if line and line[0] == '0':
                        oldline = oldline.rstrip('\r\n') + line[3:].rstrip()
                    else:
                        self._parse_line(oldline)
                        self.start_line = lineno
                        oldline = line
                else:
                    self.start_line = lineno
                    oldline = line

            self._parse_line(oldline)
        except Exception as e:
            raise SyntaxError(f'Error parsing {self.filename}:{lineno}: {e}') from e

        return Netlist(self.test_records, self.conductors, self.outlines, self.comments, self.adjacency,
                params=self.params, import_settings=self.settings, original_path=path,
                generator_hints=self.generator_hints)

    def _parse_line(self, line):
        if not line:
            return

        if self.eof:
            self.warn('Data following IPC-356 End Of File marker')

        if line[0] == 'C':
            line = line[2:].strip()
            #     +-- sic!
            #     v
            if 'Ouptut' in line and 'Allegro' in line:
                self.generator_hints.append('allegro')

            elif 'Ouptut' not in line and 'Allegro' in line:
                self.warn('This seems to be a file generated by a newer allegro version. Please raise an issue on our '
                          'issue tracker with your Allegro version and if possible please provide an example file '
                          'so we can improve Gerbonara!')

            elif 'EAGLE' in line and 'CadSoft' in line:
                self.generator_hints.append('eagle')

            if line.strip().startswith('NNAME'):
                name, *value = line.strip().split()
                value = ' '.join(value)
                self.warn('File contains non-standard Allegro-style net name alias definitions in comments.')
                if 'allegro' in self.generator_hints:
                    # it's amazing how allegro always seems to have found a way to do the same thing everyone else is
                    # doing just in a different, slightly more messed up, completely incompatible way.
                    self.net_names[name] = value[5:] # strip NNAME because Allegro

                else:
                    self.net_names[name] = value

            else:
                self.comments.append(line)

        elif line[0] == 'P':
            # Parameter
            name, *value = line[2:].split()
            value = ' '.join(value)

            if name == 'UNITS':
                if value in ('CUST', 'CUST 0'):
                    self.settings.unit = Inch
                    self.settings.angle_unit = 'degree'
                    self.has_unit = True

                elif value == 'CUST 1':
                    self.settings.unit = MM
                    self.settings.angle_unit = 'degree'
                    self.has_unit = True

                elif value == 'CUST 2':
                    self.settings.unit = Inch
                    self.settings.angle_unit = 'radian'
                    self.has_unit = True

                else:
                    raise SyntaxError(f'Unsupported IPC-356 netlist unit specification "{line}"')

            elif name.startswith('NNAME'):
                if 'allegro' in self.generator_hints:
                    self.net_names[name] = value[5:]

                else:
                    self.net_names[name] = value

            else:
                self.params[name] = value

        elif line[0] == '9':
            self.eof = True

        elif line[0:3] in ('317', '327', '367'):
            self.assert_unit()
            self.test_records.append(TestRecord.parse(line, self.settings, self.net_names))

        elif line[0:3] == '378':
            self.assert_unit()
            self.conductors.append(Conductor.parse(line, self.settings, self.net_names))

        elif line[0:3] == '379':
            net, *adjacent = line[3:].strip().split()

            for other in adjacent:
                self.adjacency[net] = self.adjacency.get(net, set()) | {other}
                self.adjacency[other] = self.adjacency.get(other, set()) | {net}

        elif line[0:3] == '389':
            self.assert_unit()
            self.outlines.extend(Outline.parse(line, self.settings))

        else:
            self.warn(f'Unknown IPC-356 record type {line[0:3]}')


class PadType(Enum):
    THROUGH_HOLE = 1
    SMD_PAD = 2
    TOOLING_FEATURE = 3
    TOOLING_HOLE = 4
    NONPLATED_HOLE = 6


class SoldermaskInfo(Enum):
    NONE = 0
    PRIMARY = 1
    SECONDARY = 2
    BOTH = 3


@dataclass
class TestRecord:
    __test__ = False # tell pytest to ignore this class
    pad_type : PadType = None
    net_name : str = None
    is_connected : bool = True # None, True or False.
    ref_des : str = None # part reference designator, e.g. "C1" or "U69"
    is_via : bool = False
    pin_num : int = None
    is_middle : bool = False # is this a point in the middle or at the end of a trace/net?
    hole_dia : float = None
    is_plated : bool = None # None, True, or False.
    access_layer : int = None
    x : float = None
    y : float = None
    w : float = None
    h : float = None
    rotation : float = 0
    solder_mask : SoldermaskInfo = None
    lefover : str = None
    unit : LengthUnit = None

    def __str__(self):
        x = self.unit.format(self.x)
        y = self.unit.format(self.y)
        return f'<IPC-356 test record @ {x},{y} {self.net_name} {self.pad_type.name} at {self.ref_des}, pin {self.pin_num}>'

    def rotate(self, angle, cx=0, cy=0, unit=None):
        cx = self.unit(cx, unit)
        cy = self.unit(cy, unit)

        self.angle += angle
        self.x, self.y = rotate_point(self.x, self.y, angle, center=(cx, cy))

    def offset(self, dx=0, dy=0, unit=None):
        dx = self.unit(dx, unit)
        dy = self.unit(dy, unit)
        self.x += dx
        self.y += dy

    @classmethod
    def parse(kls, line, settings, net_name_map={}):
        obj = kls()
        line = f'{line:<80}'

        obj.unit = settings.unit
        obj.pad_type = PadType(int(line[1]))

        net_name = line[3:17].strip() or None
        if net_name == 'N/C':
            obj.net_name = None
            obj.is_connected = False
        else:
            obj.net_name = net_name_map.get(net_name, net_name)
            obj.is_connected = True

        ref_des = line[20:26].strip() or None
        if ref_des == 'VIA':
            obj.is_via = True
            obj.ref_des = None
        else:
            obj.is_via = False
            obj.ref_des = ref_des

        obj.pin = line[27:31].strip() or None

        if line[31] == 'M':
            obj.is_middle = True
        if line[32] == 'D':
            obj.hole_dia = settings.parse_ipc_length(line[33:37])
        if line[37] in ('P', 'U'):
            obj.is_plated = (line[37] == 'P')
        if line[38] == 'A':
            obj.access_layer = int(line[39:41])
        if line[41] == 'X':
            obj.x = settings.parse_ipc_length(line[42:49])
        if line[49] == 'Y':
            obj.y = settings.parse_ipc_length(line[50:57])
        if line[57] == 'X':
            obj.w = settings.parse_ipc_length(line[58:62])
        if line[62] == 'Y':
            obj.h = settings.parse_ipc_length(line[63:67])
        if line[67] == 'R':
            obj.rotation = math.radians(int(line[68:71]))
        else:
            obj.rotation = 0
        if line[72] == 'S':
            obj.solder_mask = SoldermaskInfo(int(line[73]))
        obj.leftover = line[74:].strip() or None

        return obj

    def format(self, settings, net_name_map={}):
        x = settings.unit(self.x, self.unit)
        y = settings.unit(self.y, self.unit)
        w = settings.unit(self.w, self.unit)
        h = settings.unit(self.h, self.unit)
        # TODO: raise warning if any string is too long
        ref_des = 'VIA' if self.is_via else (self.ref_des or '')
        if self.is_connected:
            net_name = net_name_map.get(self.net_name, self.net_name)
        else:
            net_name = 'N/C'

        yield ''.join((
            '3',
            str(self.pad_type.value),
            '7',
            f'{net_name or "":<14}'[:14],
            '   ',
            f'{ref_des or "":<6}'[:6],
            '-',
            f'{self.pin_num or "":<4}'[:4],
            'M' if self.is_middle else ' ',
            settings.format_ipc_length(self.hole_dia, 4, 'D', self.unit),
            {True: 'P', False: 'U', None: ' '}[self.is_plated],
            settings.format_ipc_number(self.access_layer, 2, 'A'),
            settings.format_ipc_length(self.x, 6, 'X', self.unit, sign=True),
            settings.format_ipc_length(self.y, 6, 'Y', self.unit, sign=True),
            settings.format_ipc_length(self.w, 4, 'X', self.unit),
            settings.format_ipc_length(self.h, 4, 'Y', self.unit),
            settings.format_ipc_number(math.degrees(self.rotation) if self.rotation is not None else None, 3, 'R'),
            ' ',
            settings.format_ipc_number(self.solder_mask, 1, 'S'),
            f'{self.leftover or "":<6}'))

class OutlineType(Enum):
    BOARD_EDGE = 0
    PANEL_EDGE = 1
    SCORE_LINE = 2
    OTHER_FAB = 3


def parse_coord_chain(line, settings):
    x, y = None, None
    for segment in line.split('*'):
        coords = []
        for coord in segment.strip().split():
            if not (match := re.match(r'(X[+-]?[0-9]+)?(Y[+-]?[0-9]+)?', coord)):
                raise SyntaxError(f'Invalid IPC-356 coordinate {coord}')

            x = settings.parse_ipc_length(match[1], x)
            y = settings.parse_ipc_length(match[2], y)

            if x is None or y is None:
                raise SyntaxError('Outline or conductor coordinate chain is missing one coordinate in the beginning')

            coords.append((x, y))
        yield coords

def format_coord_chain(line, settings, coords, cont, unit):
    for x, y in coords:
        coord = settings.format_ipc_length(x, 6, 'X', unit=unit, sign=True)
        coord += settings.format_ipc_length(y, 6, 'Y', unit=unit, sign=True)

        if len(line) + len(coord) <= 80:
            line = (line + coord + ' ')[:80]

        else:
            yield line
            line = f'{cont} {coord} '
    yield line


@dataclass
class Outline:
    outline_type : OutlineType
    outline : [(float,)]
    unit : LengthUnit = None

    @classmethod
    def parse(kls, line, settings):
        outline_type = OutlineType[line[3:17].strip()]
        for outline in parse_coord_chain(line[22:], settings):
            yield kls(outline_type, outline, unit=settings.unit)

    def format(self, settings):
        line = f'389{self.outline_type.name:<14}     '
        yield from format_coord_chain(line, settings, self.outline, '089', self.unit)

    def __str__(self):
        return f'<IPC-356 {self.outline_type.name} outline with {len(self.outline)} points>'

    def rotate(self, angle, cx=0, cy=0, unit=None):
        cx = self.unit(cx, unit)
        cy = self.unit(cy, unit)
        self.outline = [ rotate_point(x, y, angle, center=(cx, cy)) for x, y in self.outline ]

    def offset(self, dx=0, dy=0, unit=None):
        dx = self.unit(dx, unit)
        dy = self.unit(dy, unit)
        self.outline = [ (x+dx, y+dy) for x, y in self.outline ]


@dataclass
class Conductor:
    net_name : str
    layer : int
    aperture : (float,)
    coords : [(float,)]
    unit : LengthUnit = None

    @classmethod
    def parse(kls, line, settings, net_name_map={}):
        net_name = line[3:17].strip() or None
        net_name = net_name_map.get(net_name, net_name)

        if line[18] != 'L':
            raise SytaxError(f'Invalid IPC-356 layer number specification for conductor in line "{line}"')
        layer = int(line[19:21])

        aperture_def, _, coords = line[22:].partition(' ')
        if not (m := re.match(r'(X[+-]?[0-9]+)(Y[+-]?[0-9]+)?', coord)):
            raise SyntaxError('Invalid IPC-356 aperture specification "{aperture_def"}')
        aperture = settings.parse_ipc_length(m[1]), settings.parse_ipc_length(m[2])

        for chain in parse_coord_chain(coords, settings):
            yield kls(net_name, layer, aperture, chain, unit=settings.unit)

    def format(self, settings, net_name_map):
        net_name = net_name_map.get(self.net_name, self.net_name)
        net_name = f'{net_name:<14}[:14]'
        line = f'378{net_name} L{self.layer:02d} '
        yield from format_coord_chain(line, settings, self.outline, '078', self.unit)

    def __str__(self):
        return f'<IPC-356 conductor {self.net_name} with {len(self.coords)} points>'

    def rotate(self, angle, cx=0, cy=0, unit=None):
        cx = self.unit(cx, unit)
        cy = self.unit(cy, unit)
        self.coords = [ rotate_point(x, y, angle, center=(cx, cy)) for x, y in self.coords ]

    def offset(self, dx=0, dy=0, unit=None):
        dx = self.unit(dx, unit)
        dy = self.unit(dy, unit)
        self.coords = [ (x+dx, y+dy) for x, y in self.coords ]




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2022 Jan Sebastian Götte <gerbonara@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import math
import itertools

from dataclasses import dataclass, replace, field

from .utils import *

prec = lambda x: f'{float(x):.6}'


@dataclass(frozen=True)
class GraphicPrimitive:

    # hackety hack: Work around python < 3.10 not having dataclasses.KW_ONLY.
    # 
    # For details, refer to graphic_objects.py
    def __init_subclass__(cls):
        cls.polarity_dark = True

        d = {'polarity_dark': bool}
        if hasattr(cls, '__annotations__'):
            cls.__annotations__.update(d)
        else:
            cls.__annotations__ = d

    def bounding_box(self):
        """ Return the axis-aligned bounding box of this feature.

        :returns: ``((min_x, min_Y), (max_x, max_y))``
        :rtype: tuple
        """

        raise NotImplementedError()

    def to_svg(self, fg='black', bg='white', tag=Tag):
        """ Render this primitive into its SVG representation.

        :param str fg: Foreground color. Must be an SVG color name.
        :param str bg: Background color. Must be an SVG color name.
        :param function tag: Tag constructor to use.

        :rtype: str
        """

        raise NotImplementedError()

    def is_zero_size(self):
        """ Return whether this primitive is zero size

        :rtype: bool
        """


@dataclass(frozen=True)
class Circle(GraphicPrimitive):
    #: Center X coordinate
    x : float
    #: Center y coordinate
    y : float
    #: Radius, not diameter like in :py:class:`.apertures.CircleAperture`
    r : float # Here, we use radius as common in modern computer graphics, not diameter as gerber uses.

    def bounding_box(self):
        return ((self.x-self.r, self.y-self.r), (self.x+self.r, self.y+self.r))

    def to_svg(self, fg='black', bg='white', tag=Tag):
        color = fg if self.polarity_dark else bg
        return tag('circle', cx=prec(self.x), cy=prec(self.y), r=prec(self.r), fill=color)

    def to_arc_poly(self):
        return ArcPoly([(self.x-self.r, self.y), (self.x+self.r, self.y)],
                       [(True, (self.x, self.y)), (True, (self.x, self.y))],
                       polarity_dark=self.polarity_dark)

    def is_zero_size(self):
        return math.isclose(self.r, 0)


@dataclass(frozen=True)
class ArcPoly(GraphicPrimitive):
    """ Polygon whose sides may be either straight lines or circular arcs. """

    #: list of (x : float, y : float) tuples. Describes closed outline, i.e. the first and last point are considered
    #: connected.
    outline : list
    #: Must be either None (all segments are straight lines) or same length as outline.
    #: Straight line segments have None entry. Arc segments have (clockwise, (cx, cy)) tuple with cx, cy being absolute
    #: coords.
    arc_centers : list = field(default_factory=list)

    @property
    def segments(self):
        """ Return an iterator through all *segments* of this polygon. For each outline segment (line or arc), this
        iterator will yield a ``(p1, p2, (clockwise, center))`` tuple. If the segment is a straight line, ``clockwise``
        will be ``None``.
        """
        for points, arc in itertools.zip_longest(itertools.pairwise(self.outline), self.arc_centers):
            if arc:
                if points:
                    yield *points, arc
                else:
                    yield self.outline[-1], self.outline[0], arc
                    return
            else:
                if not points:
                    break
                yield *points, (None, (None, None))

        # Close outline if necessary.
        if math.dist(self.outline[0], self.outline[-1]) > 1e-6:
            yield self.outline[-1], self.outline[0], (None, (None, None))

    def approximate_arcs(self, max_error=1e-2, clip_max_error=True):
        outline = []
        for (x1, y1), (x2, y2), (clockwise, (cx, cy)) in self.segments:
            if clockwise is None:
                outline.append((x1, y1))
            else:
                outline.extend(approximate_arc(cx, cy, x1, y1, x2, y2, clockwise,
                                               max_error=max_error, clip_max_error=clip_max_error))
                outline.pop() # remove arc end point
        return type(self)(outline, polarity_dark=self.polarity_dark)

    def bounding_box(self):
        bbox = (None, None), (None, None)
        for (x1, y1), (x2, y2), (clockwise, (cx, cy)) in self.segments:
            if clockwise is None:
                line_bounds = (min(x1, x2), min(y1, y2)), (max(x1, x2), max(y1, y2))
                bbox = add_bounds(bbox, line_bounds)
            else:
                bbox = add_bounds(bbox, arc_bounds(x1, y1, x2, y2, cx, cy, clockwise))
        return bbox

    @classmethod
    def from_regular_polygon(kls, x:float, y:float, r:float, n:int, rotation:float=0, polarity_dark:bool=True):
        """ Convert an n-sided gerber polygon to a normal ArcPoly defined by outline """

        delta = 2*math.pi / n

        return kls([
                (x + math.cos(rotation + i*delta) * r,
                 y + math.sin(rotation + i*delta) * r)
                for i in range(n) ], polarity_dark=polarity_dark)

    def __len__(self):
        """ Return the number of points on this polygon's outline (which is also the number of segments because the
        polygon is closed). """
        return len(self.outline)

    def __bool__(self):
        """ Return ``True`` if this polygon has any outline points. """
        return bool(len(self))

    def path_d(self):
        if len(self.outline) == 0:
            return

        yield f'M {float(self.outline[0][0]):.6} {float(self.outline[0][1]):.6}'

        for old, new, (clockwise, center) in self.segments:
            if clockwise is None:
                yield f'L {float(new[0]):.6} {float(new[1]):.6}'
            else:
                yield svg_arc(old, new, center, clockwise)

    def to_svg(self, fg='black', bg='white', tag=Tag):
        color = fg if self.polarity_dark else bg
        return tag('path', d=' '.join(self.path_d()), fill=color)

    def to_arc_poly(self):
        return self


    def is_zero_size(self):
        for (x1, y1), (x2, y2), (clockwise, (cx, cy)) in self.segments:
            if clockwise is not None: # arc
                if math.isclose(cx, x1) and math.isclose(cy, y1):
                    continue
                
                if math.isclose(x1, x2) and math.isclose(y1, y2):
                    return False

        if math.isclose(polygon_area(self.outline), 0):
            return True
        return False


@dataclass(frozen=True)
class Line(GraphicPrimitive):
    """ Straight line with round end caps. """
    #: Start X coordinate. As usual in modern graphics APIs, this is at the center of the half-circle capping off this
    #: line.
    x1 : float
    #: Start Y coordinate
    y1 : float
    #: End X coordinate
    x2 : float
    #: End Y coordinate
    y2 : float
    #: Line width
    width : float

    def flip(self):
        return replace(self, x1=self.x2, y1=self.y2, x2=self.x1, y2=self.y1)

    @classmethod
    def from_obround(kls, x:float, y:float, w:float, h:float, rotation:float=0, polarity_dark:bool=True):
        """ Convert a gerber obround into a :py:class:`~.graphic_primitives.Line`. """
        if w > h:
            w, a, b = h, w-h, 0
        else:
            w, a, b = w, 0, h-w

        return kls(
                *rotate_point(x-a/2, y-b/2, rotation, x, y),
                *rotate_point(x+a/2, y+b/2, rotation, x, y),
                w, polarity_dark=polarity_dark)

    def bounding_box(self):
        r = self.width / 2
        return add_bounds(Circle(self.x1, self.y1, r).bounding_box(), Circle(self.x2, self.y2, r).bounding_box())

    def to_svg(self, fg='black', bg='white', tag=Tag):
        color = fg if self.polarity_dark else bg
        width = f'{self.width:.6}' if not math.isclose(self.width, 0) else '0.01mm'
        return tag('path', d=f'M {float(self.x1):.6} {float(self.y1):.6} L {float(self.x2):.6} {float(self.y2):.6}',
                fill='none', stroke=color, stroke_width=str(width), stroke_linecap='round')

    def to_arc_poly(self):
        l = math.dist((self.x1, self.y1), (self.x2, self.y2))
        if math.isclose(l, 0):
            # degenerate case: a zero-length line becomes a circle.
            return ArcPoly([(self.x1-self.width/2, self.y1), (self.x1+self.width/2, self.y1)],
                           [(True, (self.x1, self.y1)), (True, (self.x1, self.y1))],
                           polarity_dark=self.polarity_dark)

        dx, dy = self.x2-self.x1, self.y2-self.y1
        nx, ny = -dy/l, dx/l
        rx, ry = nx*self.width/2, ny*self.width/2
        return ArcPoly([
                    (self.x2+rx, self.y2+ry),
                    (self.x2-rx, self.y2-ry),
                    (self.x1-rx, self.y1-ry),
                    (self.x1+rx, self.y1+ry),
                ], [
                    (True, (self.x2, self.y2)),
                    None,
                    (True, (self.x1, self.y1)),
                    None,
                ], polarity_dark=self.polarity_dark)

    def is_zero_size(self):
        return math.isclose(self.x1, self.x2) and math.isclose(self.y1, self.y2)


@dataclass(frozen=True)
class Arc(GraphicPrimitive):
    """ Circular arc with line width ``width`` going from ``(x1, y1)`` to ``(x2, y2)`` around center at ``(cx, cy)``. """
    #: Start X coodinate
    x1 : float
    #: Start Y coodinate
    y1 : float
    #: End X coodinate
    x2 : float
    #: End Y coodinate
    y2 : float
    #: Center X coordinate (absolute)
    cx : float
    #: Center Y coordinate (absolute)
    cy : float
    #: ``True`` if this arc is clockwise from start to end. Selects between the large arc and the small arc given this
    #: start, end and center
    clockwise : bool
    #: Line width of this arc.
    width : float

    @property
    def is_circle(self):
        return math.isclose(self.x1, self.x2, abs_tol=1e-6) and math.isclose(self.y1, self.y2, abs_tol=1e-6)

    def flip(self):
        return replace(self, x1=self.x2, y1=self.y2, x2=self.x1, y2=self.y1, clockwise=not self.clockwise)

    def bounding_box(self):
        r = self.width/2
        (min_x, min_y), (max_x, max_y) = arc_bounds(self.x1, self.y1, self.x2, self.y2, self.cx, self.cy, self.clockwise)
        return (min_x-r, min_y-r), (max_x+r, max_y+r)

    def to_svg(self, fg='black', bg='white', tag=Tag):
        color = fg if self.polarity_dark else bg
        arc = svg_arc((self.x1, self.y1), (self.x2, self.y2), (self.cx, self.cy), self.clockwise)
        width = f'{self.width:.6}' if not math.isclose(self.width, 0) else '0.01mm'
        return tag('path', d=f'M {float(self.x1):.6} {float(self.y1):.6} {arc}',
                fill='none', stroke=color, stroke_width=width, stroke_linecap='round')

    def to_arc_poly(self):
        r = math.dist((self.x1, self.y1), (self.cx, self.cy))

        if math.isclose(r, 0):
            # degenerate case: a zero-radius arc becomes a circle.
            return ArcPoly([(self.x1-self.width/2, self.y1), (self.x1+self.width/2, self.y1)],
                           [(True, (self.x1, self.y1)), (True, (self.x1, self.y1))],
                           polarity_dark=self.polarity_dark)

        dx1, dy1 = self.x1-self.cx, self.y1-self.cy
        nx1, ny1 = dx1/r * self.width/2, dy1/r * self.width/2
        dx2, dy2 = self.x2-self.cx, self.y2-self.cy
        nx2, ny2 = dx2/r * self.width/2, dy2/r * self.width/2
        return ArcPoly([ # vertices
                    (self.x1+nx1, self.y1+ny1),
                    (self.x1-nx1, self.y1-ny1),
                    (self.x2-nx2, self.y2-ny2),
                    (self.x2+nx2, self.y2+ny2),
                ], [ # arc segments (direction, center)
                    (not self.clockwise, (self.x1, self.y1)),
                    (self.clockwise, (self.cx, self.cy)),
                    (self.clockwise, (self.x2, self.y2)),
                    (not self.clockwise, (self.cx, self.cy)),
                ], polarity_dark=self.polarity_dark)

    def is_zero_size(self):
        return False # an arc with identical start and end points is defined as a circle


@dataclass(frozen=True)
class Rectangle(GraphicPrimitive):
    #: **Center** X coordinate
    x : float
    #: **Center** Y coordinate
    y : float
    #: width
    w : float
    #: height
    h : float
    #: rotation around center in radians
    rotation : float

    def bounding_box(self):
        return self.to_arc_poly().bounding_box()

    def to_arc_poly(self):
        sin, cos = math.sin(self.rotation), math.cos(self.rotation)
        sw, cw = sin*self.w/2, cos*self.w/2
        sh, ch = sin*self.h/2, cos*self.h/2
        x, y = self.x, self.y
        return ArcPoly([
            (x - (cw+sh), y - (ch+sw)),
            (x - (cw+sh), y + (ch+sw)),
            (x + (cw+sh), y + (ch+sw)),
            (x + (cw+sh), y - (ch+sw)),
            ], polarity_dark=self.polarity_dark)

    def to_svg(self, fg='black', bg='white', tag=Tag):
        color = fg if self.polarity_dark else bg
        x, y = self.x - self.w/2, self.y - self.h/2
        return tag('rect', x=prec(x), y=prec(y), width=prec(self.w), height=prec(self.h),
                **svg_rotation(self.rotation, self.x, self.y), fill=color)

    def is_zero_size(self):
        return math.isclose(self.w, 0) or math.isclose(self.h, 0)

